# Copyright 2019 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import time
import logging
from collections import OrderedDict

from chirp import chirp_common, directory, memmap, errors, util
from chirp import bitwise
from chirp.drivers import tk280
from chirp.settings import MemSetting, RadioSettingValueInvertedBoolean
from chirp.settings import RadioSettingGroup, RadioSetting
from chirp.settings import RadioSettingValueBoolean, RadioSettingValueList
from chirp.settings import RadioSettingValueString, RadioSettingValueInteger
from chirp.settings import RadioSettings, RadioSettingSubGroup
from chirp.settings import RadioSettingValueMap

LOG = logging.getLogger(__name__)

HEADER_FORMAT = """
#seekto 0x0100;
struct {
  char sw_name[7];
  char sw_ver[5];
  u8 unknown1[4];
  char sw_key[12];
  u8 unknown2[4];
  char model[5];
  u8 variant;
  u8 unknown3[10];
} header;

#seekto 0x0140;
struct {
  // 0x0140
  u8 unknown1;
  u8 sublcd;
  u8 unknown2[30];

  // 0x0160
  char pon_msgtext[12];
  u8 min_volume;
  u8 max_volume;
  u8 lo_volume;
  u8 hi_volume;

  // 0x0170
  u8 tone_volume_offset;
  u8 poweron_tone;
  u8 control_tone;
  u8 warning_tone;
  u8 alert_tone;
  u8 sidetone;
  u8 locator_tone;
  u8 battsave; // FF=off, 30=short, 31=med, 32=long
  u8 battwarn; // FF=off, 30=Transmit, 31=Always, 32=AlwaysBeep
  u8 ignition_mode;
  u8 ignition_time;  // In tens of minutes (6 = 1h)
  u8 micsense;
  ul16 modereset;
  u8 min_vol_preset;
  u8 unknown4;

  // 0x0180
  u8 pttid_type; // 30=DTMF, 31=FleetSync
  u8 pttid_dtmf_bot_count; // number of digits
  u8 pttid_dtmf_bot_code[8]; // digits, 0xFF padded
  u8 pttid_dtmf_eot_count; // number of digits
  u8 pttid_dtmf_eot_code[8]; // digits, 0xFF padded

  // 0x0193
  u8 pon_msgtype;
  u8 unknown7[8];
  u8 battstatus:1,
     unknown8_1:1,
     ssi:1,
     busy_led:1,
     power_switch_memory:1,
     scrambler_memory:1,
     unknown8_2:1,
     off_hook_decode:1;
  u8 unknown9_1:5,
     clockfmt:1,
     datefmt:1,
     ignition_sense:1;
  u8 unknownA[2];

  // 0x01A0
  u8 unknownB[8];
  u8 ptt_timer;
  u8 unknownB2[3];
  u8 ptt_proceed:1,
     unknownC_1:3,
     tone_off:1,
     ost_memory:1,
     unknownC_2:1,
     ptt_release:1;
  u8 unknownD[3];
} settings;

#seekto 0x01E0;
struct {
  char name[12];
  ul16 rxtone;
  ul16 txtone;
} ost_tones[40];

// Button config
// 0xFF = none 0x00=2t 0x26=lamp 04=autotel 01=autodial 12= 5=chentry
// 13=clock 02=autodial-prog 08=call1 09=call2 (up to call6) 21=function
// 20=fixedvol 28=lowpow 2A=monitor 2B=monmoment 42=TA 43=teldisc
// 1E=displaychar 3D=sqlvl 3E=sqoff 3F=sqoffmom 32=password
// 10=grpup (but more changes)
// 0E=grpdn (but more changes)
// 0x582 selector mode FF=none 30=group 31=channel
// 0x583 keypad type ff=none 30=dtmf
// 0x584 keypad op:
// 0x58E 0x04 list sel key (inverted)
// ff=none
#seekto 0x0582;
u8 selector_knob; // FF=none 30=group 31=channel
u8 keypad_type; // ff=none 30=12-key
// 30=channel 31=OST 32=DTMF(autodial) 33=DTMF(keypad auto ptt)
// 34=FS(SC) 35=FS(ST) 36=FS(SC+ST)
u8 keypad_op;
u8 lsk_unknown1:5,
   list_selector_key:1,
   lsk_unknown2:3;
#seekto 0x05C0;
// side1 primary 0x5C0
// S 0x5C8
// A 0x5D0
// B 0x5D8
// C 0x5E0
// side2 primary 0x5E8
// two?
// aux primary 0x600
// 0 0x608
// 1 0x610
// .. *
// 0x668 ?
// mic pf1 0x688
// mic pf2 0x690

struct button {
  u8 primary;
  u8 secondary;
  u8 unknown[6]; // These all change for some of the more advanced ones
};
struct button buttons[26];

#seekto 0x0A00;
ul16 zone_starts[128];

struct zoneinfo {
  u8 number;
  u8 zonetype;
  u8 unknown1[2];
  u8 count;
  char name[12];
  u8 unknown2[2];
  ul16 timeout;    // 15-1200
  ul16 tot_alert;  // 10
  ul16 tot_rekey;  // 60
  ul16 tot_reset;  // 15
  u8 unknown3[3];
  u8 unknown21:2,
     bcl_override:1,
     unknown22:5;
  u8 unknown5;
};

struct memory {
  u8 number;
  lbcd rx_freq[4];
  lbcd tx_freq[4];
  u8 unknown11:4,
     rx_step:4;
  u8 unknown12:4,
     tx_step:4;
  ul16 rx_tone;
  ul16 tx_tone;
  char name[12];
  u8 unknown1[2];
  u8 pttid;
  u8 unknown2[16];
  u8 unknown3_1:4,
     highpower:1,
     unknown3_2:1,
     wide:1,
     unknown3_3:1;
  u8 unknown4;
};

#seekto 0xC570;  // Fixme
u8 skipflags[64];
"""


SYSTEM_MEM_FORMAT = """
#seekto 0x%(addr)x;
struct {
  struct zoneinfo zoneinfo;
  struct memory memories[%(count)i];
} zone%(index)i;
"""

STARTUP_MODES = ['Text', 'Clock']

VOLUMES = OrderedDict([(str(x), x) for x in range(0, 30)])
VOLUMES.update({'Selectable': 0x30,
                'Current': 0xFF})
VOLUMES_REV = {v: k for k, v in VOLUMES.items()}

MIN_VOL_PRESET = {'Preset': 0x30,
                  'Lowest Limit': 0x31}
MIN_VOL_PRESET_REV = {v: k for k, v in MIN_VOL_PRESET.items()}

SUBLCD = ['Zone Number', 'CH/GID Number', 'OSD List Number']
CLOCKFMT = ['12H', '24H']
DATEFMT = ['Day/Month', 'Month/Day']
MICSENSE = ['On']
ONLY_MOBILE_SETTINGS = ['power_switch_memory', 'off_hook_decode',
                        'ignition_sense', 'mvp', 'it', 'ignition_mode']
PTTID_SETTINGS = {
    'Off': 0xFF,
    'BOT': 0x30,
    'EOT': 0x31,
    'Both': 0x32,
}
PTTID_TYPES = {
    'DTMF': 0x30,
    'FleetSync': 0x31,
}
BATTSAVE_SETTINGS = {
    'Off': 0xFF,
    'Short': 0x30,
    'Medium': 0x31,
    'Long': 0x32,
}
BATTWARN_SETTINGS = {
    'Off': 0xFF,
    'Transmit': 0x30,
    'Always': 0x31,
    'Always with Beep': 0x32,
}
# The buttons in memory are not logically arranged for display and sorting
# would do weird things, so map them manually here
BUTTONS = {
    'Side 1 / Triangle': 0,
    'Side 2': 5,
    'S': 1,
    'A': 2,
    'B': 3,
    'C': 4,
    # Skip 2
    'Top Aux': 8,
    'DTMF 0': 9,
    'DTMF 1': 10,
    'DTMF 2': 11,
    'DTMF 3': 12,
    'DTMF 4': 13,
    'DTMF 5': 14,
    'DTMF 6': 15,
    'DTMF 7': 16,
    'DTMF 8': 17,
    'DTMF 9': 18,
    'DTMF *': 19,
    'DTMF #': 20,
    'Mic PF1': 24,
    'Mic PF2': 25,
}
# IDs of the buttons that can only be primary
PRIMARY_ONLY = [0x0E, 0x10, 0x21, 0x2B, 0x3F, 0x4B, 0x4C]
BUTTON_FUNCTIONS = {
    'None': 0xFF,
    '2-tone': 0x00,
    'Autodial': 0x01,
    'Autodial Programming': 0x02,
    # 0x03
    'Auto Telephone': 0x04,
    # 0x06 0x07
    'Call 1': 0x08,
    'Call 2': 0x09,
    'Call 3': 0x0A,
    'Call 4': 0x0B,
    'Call 5': 0x0C,
    'Call 6': 0x0D,
    # 0x0D
    'CH/GID Down': 0x0E,
    # 0x0F
    'CH/GID Up': 0x10,
    # 0x11
    'Channel Entry': 0x12,
    'Clock': 0x13,
    'Display Character': 0x1E,
    'Fixed Volume': 0x20,
    'Function': 0x21,
    'LCD Brightness': 0x27,
    'Monitor': 0x2A,
    'Monitor Momentary': 0x2B,
    'Transceiver Password': 0x32,
    'Squelch Level': 0x3D,
    'Squelch Off': 0x3E,
    'Squelch Off (momentary)': 0x3F,
    'Talk Around': 0x42,
    'Telephone Disconnect': 0x43,
    'Zone Down': 0x4B,
    'Zone Up': 0x4C,
    'CH/GID Recall': 0x4E,
}
# These are only available on the portable transceiver
PORTABLE_BUTTON_FUNCTIONS = {
    'AUX': 0x05,
    'Lamp': 0x26,
    'Low TX Power': 0x28,
}
# These are only available on the mobile transceiver
MOBILE_BUTTON_FUNCTIONS = {
    'Zone Down': 0x4A,  # Different on mobile?
}
KNOB_MODE = {
    'None': 0xFF,
    'Group': 0x30,
    'Channel': 0x31,
}
KEYPAD_TYPE = {
    'None': 0xFF,
    '12-key': 0x30,
}
KEYPAD_OP = {
    'None': 0xFF,
    'Channel': 0x30,
    'OST': 0x31,
    'DTMF (autodial)': 0x32,
    'DTMF (Keypad Auto PTT)': 0x33,
    'FleetSync (Selcall)': 0x34,
    'FleetSync (Status)': 0x35,
    'FleetSync (Selcall+Status)': 0x36,
}

POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5),
                chirp_common.PowerLevel("High", watts=50)]


def decode_dtmf(array):
    return ''.join('%x%x' % (int(i) & 0x0F, (int(i) & 0xF0) >> 4)
                   for i in array if i != 0xFF)


def encode_dtmf(string):
    nibs = [int(c, 16) for c in string]
    nibs += [0xF for i in range(16 - len(nibs))]
    bytevals = [(nibs[i + 1] << 4) | nibs[i] for i in range(0, 16, 2)]
    return bytevals


def reverse_dict(d):
    return {v: k for k, v in d.items()}


def set_choice(setting, obj, key, choices, default='Off'):
    settingstr = str(setting.value)
    if settingstr == default:
        val = 0xFF
    else:
        val = choices.index(settingstr) + 0x30
    setattr(obj, key, val)


def get_choice(obj, key, choices, default='Off'):
    val = getattr(obj, key)
    if val == 0xFF:
        return default
    else:
        return choices[val - 0x30]


def make_frame(cmd, addr, data=b""):
    return struct.pack(">BH", ord(cmd), addr) + data


def send(radio, frame):
    # LOG.debug("%04i P>R:\n%s" % (len(frame), util.hexprint(frame)))
    radio.pipe.write(frame)


def do_ident(radio):
    radio.pipe.baudrate = 9600
    radio.pipe.stopbits = 2
    radio.pipe.timeout = 1
    send(radio, b'PROGRAM')
    ack = radio.pipe.read(1)
    LOG.debug('Read %r from radio' % ack)
    if ack != b'\x16':
        raise errors.RadioError('Radio refused hi-speed program mode')
    radio.pipe.baudrate = 19200
    ack = radio.pipe.read(1)
    if ack != b'\x06':
        raise errors.RadioError('Radio refused program mode')
    radio.pipe.write(b'\x02')
    ident = radio.pipe.read(8)
    LOG.debug('Radio ident is %r' % ident)
    radio.pipe.write(b'\x06')
    ack = radio.pipe.read(1)
    if ack != b'\x06':
        raise errors.RadioError('Radio refused program mode')
    if ident[:6] not in (radio._model,):
        model = ident[:5].decode()
        variants = {b'\x06': 'K, K1, K3 (450-520 MHz)',
                    b'\x07': 'K2, K4 (400-470 MHz)'}
        if model == 'P3180':
            model += ' ' + variants.get(ident[5], '(Unknown)')
        raise errors.RadioError('Unsupported radio model %s' % model)


def checksum_data(data):
    _chksum = 0
    for byte in data:
        _chksum = (_chksum + byte) & 0xFF
    return _chksum


def do_download(radio):
    do_ident(radio)

    data = bytes()

    def status():
        status = chirp_common.Status()
        status.cur = len(data)
        status.max = radio._memsize
        status.msg = "Cloning from radio"
        radio.status_fn(status)

    # Addresses 0x0000-0xBF00 pulled by block number (divide by 0x100)
    for block in range(0, 0xBF + 1):
        radio.pipe.log('Read block index %02x' % block)
        send(radio, make_frame('R', block))
        cmd = radio.pipe.read(1)
        chunk = b''
        if cmd == b'Z':
            data += bytes(b'\xff' * 256)
        elif cmd == b'W':
            chunk = bytes(radio.pipe.read(256))
            if len(chunk) != 256:
                LOG.error('Received %i for block %02x' % (len(chunk), block))
                raise errors.RadioError('Radio did not send block')
            data += chunk
        else:
            LOG.error('Radio sent %r (%02x), expected W(0x57)' % (cmd,
                                                                  chr(cmd)))
            raise errors.RadioError('Radio sent unexpected response')

        status()

        chksum = radio.pipe.read(1)
        if len(chksum) != 1:
            LOG.error('Checksum was %r' % chksum)
            raise errors.RadioError('Radio sent invalid checksum')
        _chksum = checksum_data(chunk)

        if chunk and _chksum != ord(chksum):
            LOG.error(
                'Checksum failed for %i byte block 0x%02x: %02x != %02x' % (
                    len(chunk), block, _chksum, ord(chksum)))
            raise errors.RadioError('Checksum failure while reading block. '
                                    'Check serial cable.')

        radio.pipe.write(b'\x06')
        if radio.pipe.read(1) != b'\x06':
            raise errors.RadioError('Post-block exchange failed')

    # Addresses 0xC000 - 0xD1F0 pulled by address
    for block in range(0x0100, 0x1200, 0x40):
        radio.pipe.log('Read memory address %04x' % block)
        send(radio, make_frame('S', block, b'\x40'))
        x = radio.pipe.read(1)
        if x != b'X':
            raise errors.RadioError('Radio did not send block')
        chunk = radio.pipe.read(0x40)
        data += chunk

        status()

        radio.pipe.write(b'\x06')
        if radio.pipe.read(1) != b'\x06':
            raise errors.RadioError('Post-block exchange failed')

    radio.pipe.write(b'E')
    if radio.pipe.read(1) != b'\x06':
        raise errors.RadioError('Radio failed to acknowledge completion')

    LOG.debug('Read %i bytes total' % len(data))
    return data


def do_upload(radio):
    do_ident(radio)

    def status(addr):
        status = chirp_common.Status()
        status.cur = addr
        status.max = radio._memsize
        status.msg = "Cloning to radio"
        radio.status_fn(status)

    for block in range(0, 0xBF + 1):
        addr = block * 0x100
        chunk = bytes(radio._mmap[addr:addr + 0x100])
        if all(byte == b'\xff' for byte in chunk):
            send(radio, make_frame('Z', block, b'\xFF'))
        else:
            radio.pipe.log('Sending block %i' % block)
            checksum = checksum_data(chunk)
            send(radio, make_frame('W', block, chunk + bytes([checksum])))

        ack = radio.pipe.read(1)
        if ack != b'\x06':
            LOG.error('Radio refused block 0x%02x with %r' % (block, ack))
            raise errors.RadioError('Radio refused data block')

        status(addr)

    addr_base = 0xC000
    for addr in range(addr_base, radio._memsize, 0x40):
        block_addr = addr - addr_base + 0x0100
        chunk = radio._mmap[addr:addr + 0x40]
        send(radio, make_frame('X', block_addr, b'\x40' + chunk))

        ack = radio.pipe.read(1)
        if ack != b'\x06':
            LOG.error('Radio refused address 0x%02x with %r' % (block_addr,
                                                                ack))
            raise errors.RadioError('Radio refused data block')

        status(addr)

    radio.pipe.write(b'E')
    if radio.pipe.read(1) != b'\x06':
        raise errors.RadioError('Radio failed to acknowledge completion')


def reset(self):
    try:
        self.pipe.baudrate = 9600
        self.pipe.write(b'E')
        time.sleep(0.5)
        self.pipe.baudrate = 19200
        self.pipe.write(b'E')
    except Exception:
        LOG.error('Unable to send reset sequence')


class KenwoodTKx180Radio(chirp_common.CloneModeRadio):
    """Kenwood TK-x180"""
    VENDOR = 'Kenwood'
    MODEL = 'TK-x180'
    BAUD_RATE = 9600
    FORMATS = [directory.register_format('Kenwood KPG-89D', '*.dat')]

    _system_start = 0x0B00
    _memsize = 0xD100

    @property
    def is_mobile(self):
        """True if this is a mobile transceiver"""
        return self.MODEL[3] in ('78')

    def __init__(self, *a, **k):
        self._zones = []
        chirp_common.CloneModeRadio.__init__(self, *a, **k)
        self._dat_header = (b'KPG89D\xFF\xFF\xFF\xFFV1.61' + self._model +
                            (b'\xFF' * 11) + (b'\xFF' * 32))

    def sync_in(self):
        try:
            data = do_download(self)
            self._mmap = memmap.MemoryMapBytes(data)
        except errors.RadioError:
            reset(self)
            raise
        except Exception as e:
            reset(self)
            LOG.exception('General failure')
            raise errors.RadioError('Failed to download from radio: %s' % e)
        self.process_mmap()

    def sync_out(self):
        try:
            do_upload(self)
        except Exception as e:
            reset(self)
            LOG.exception('General failure')
            raise errors.RadioError('Failed to upload to radio: %s' % e)

    @classmethod
    def match_model(cls, filedata, filename):
        return (filename.endswith('.dat') and
                filedata[15:21] == cls._model)

    def save_mmap(self, filename):
        if filename.lower().endswith('.dat'):
            with open(filename, 'wb') as f:
                f.write(self._dat_header)
                f.write(self._mmap.get_packed())
                LOG.info('Wrote DAT file')
        else:
            super().save_mmap(filename)

    def load_mmap(self, filename):
        if filename.lower().endswith('.dat'):
            with open(filename, 'rb') as f:
                self._dat_header = f.read(0x40)
                LOG.debug('DAT header:\n%s' % util.hexprint(self._dat_header))
                self._mmap = memmap.MemoryMapBytes(f.read())
                LOG.info('Loaded DAT file')
            self.process_mmap()
        else:
            super().load_mmap(filename)

    @property
    def is_portable(self):
        return self._model.startswith(b'P')

    def probe_layout(self):
        tmp_format = '#seekto 0x0A00; ul16 zone_starts[128];'
        mem = bitwise.parse(tmp_format, self._mmap)
        zone_format = """struct zoneinfo {
        u8 number;
        u8 zonetype;
        u8 unknown1[2];
        u8 count;
        char name[12];
        u8 unknown2[15];
        };"""

        zone_addresses = []
        for i in range(0, 128):
            if mem.zone_starts[i] == 0xFFFF:
                break
            zone_addresses.append(mem.zone_starts[i])
            zone_format += '#seekto 0x%x; struct zoneinfo zone%i;' % (
                mem.zone_starts[i], i)

        zoneinfo = bitwise.parse(zone_format, self._mmap)
        zones = []
        for i, addr in enumerate(zone_addresses):
            zone = getattr(zoneinfo, 'zone%i' % i)
            if zone.zonetype != 0x31:
                LOG.error('Zone %i is type 0x%02x; '
                          'I only support 0x31 (conventional)')
                raise errors.RadioError(
                    'Unsupported non-conventional zone found in radio; '
                    'Refusing to load to safeguard your data!')
            zones.append((addr, zone.count))

        LOG.debug('Zones: %s' % zones)
        return zones

    def process_mmap(self):
        self._zones = self.probe_layout()

        mem_format = HEADER_FORMAT
        for index, (addr, count) in enumerate(self._zones):
            mem_format += '\n\n' + (
                SYSTEM_MEM_FORMAT % {
                    'addr': addr,
                    'count': max(count, 1),  # Never allow array of zero
                    'index': index})

        self._memobj = bitwise.parse(mem_format, self._mmap)

    def expand_mmap(self, zone_sizes):
        """Remap memory into zones of the specified sizes, copying things
        around to keep the contents, as appropriate."""
        old_zones = self._zones
        old_memobj = self._memobj

        self._mmap = memmap.MemoryMapBytes(bytes(self._mmap.get_packed()))

        new_format = HEADER_FORMAT
        addr = self._system_start
        self._zones = []
        for index, count in enumerate(zone_sizes):
            new_format += SYSTEM_MEM_FORMAT % {
                'addr': addr,
                'count': max(count, 1),  # Never allow array of zero
                'index': index}
            self._zones.append((addr, count))
            addr += 0x20 + (count * 0x30)

        self._memobj = bitwise.parse(new_format, self._mmap)

        # Set all known zone addresses and clear the rest
        for index in range(0, 128):
            try:
                self._memobj.zone_starts[index] = self._zones[index][0]
            except IndexError:
                self._memobj.zone_starts[index] = 0xFFFF

        for zone_number, count in enumerate(zone_sizes):
            dest_zone = getattr(self._memobj, 'zone%i' % zone_number)
            dest = dest_zone.memories
            dest_zoneinfo = dest_zone.zoneinfo

            if zone_number < len(old_zones):
                LOG.debug('Copying existing zone %i' % zone_number)
                _, old_count = old_zones[zone_number]
                source_zone = getattr(old_memobj, 'zone%i' % zone_number)
                source = source_zone.memories
                source_zoneinfo = source_zone.zoneinfo

                if old_count != count:
                    LOG.debug('Zone %i going from %i to %i' % (zone_number,
                                                               old_count,
                                                               count))

                # Copy the zone record from the source, but then update
                # the count
                dest_zoneinfo.set_raw(source_zoneinfo.get_raw(asbytes=False))
                dest_zoneinfo.count = count

                for dest_i in range(0, min(count, old_count)):
                    dest[dest_i].set_raw(source[dest_i].get_raw(asbytes=False))
            else:
                LOG.debug('New zone %i' % zone_number)
                dest_zone.zoneinfo.number = zone_number + 1
                dest_zone.zoneinfo.zonetype = 0x31
                dest_zone.zoneinfo.count = count
                dest_zone.zoneinfo.name = (
                    'Zone %i' % (zone_number + 1)).ljust(12)

                # Copy the settings we care about from the first zone
                z0info = self._memobj.zone0.zoneinfo
                dest_zone.zoneinfo.timeout = z0info.timeout
                dest_zone.zoneinfo.tot_alert = z0info.tot_alert
                dest_zone.zoneinfo.tot_rekey = z0info.tot_rekey
                dest_zone.zoneinfo.tot_reset = z0info.tot_reset

    def shuffle_zone(self):
        """Sort the memories in the zone according to logical channel number"""
        # FIXME: Move this to the zone
        raw_memories = self.raw_memories
        memories = [(i, raw_memories[i].number)
                    for i in range(0, self.raw_zoneinfo.count)]
        current = memories[:]
        memories.sort(key=lambda t: t[1])
        if current == memories:
            LOG.debug('Shuffle not required')
            return
        raw_data = [raw_memories[i].get_raw(asbytes=False)
                    for i, n in memories]
        for i, raw_mem in enumerate(raw_data):
            raw_memories[i].set_raw(raw_mem)

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.info = ('This radio is zone-based, which is different from how '
                   'most radios work (that CHIRP supports). The zone count '
                   'can be adjusted in the Settings tab, but you must save '
                   'and re-load the file after changing that value in order '
                   'to be able to add/edit memories there.')
        rp.experimental = ('This driver is very experimental. Every attempt '
                           'has been made to be overly pedantic to avoid '
                           'destroying data. However, you should use caution, '
                           'maintain backups, and proceed at your own risk.')
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_tuning_step = False
        rf.has_settings = True
        rf.has_bank = False
        rf.has_sub_devices = True
        rf.has_dynamic_subdevices = True
        rf.has_rx_dtcs = True
        rf.can_odd_split = True
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = ['Tone->Tone', 'DTCS->', '->DTCS', 'Tone->DTCS',
                                'DTCS->Tone', '->Tone', 'DTCS->DTCS']
        rf.valid_bands = self.VALID_BANDS
        rf.valid_modes = ['FM', 'NFM']
        rf.valid_tuning_steps = [2.5, 5.0, 6.25, 12.5, 10.0, 15.0, 20.0,
                                 25.0, 50.0, 100.0]
        rf.valid_duplexes = ['', '-', '+', 'split', 'off']
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_name_length = 12
        rf.valid_characters = ('ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                               'abcdefghijklmnopqrstuvwxyz'
                               '0123456789'
                               '!"#$%&\'()~+-,./:;<=>?@[\\]^`{}*| ')
        rf.memory_bounds = (1, 512)
        return rf

    @property
    def raw_zone(self):
        return getattr(self._memobj, 'zone%i' % self._zone)

    @property
    def raw_zoneinfo(self):
        return self.raw_zone.zoneinfo

    @property
    def raw_memories(self):
        return self.raw_zone.memories

    @property
    def max_mem(self):
        return self.raw_memories[self.raw_zoneinfo.count].number

    def _get_raw_memory(self, number):
        for i in range(0, self.raw_zoneinfo.count):
            if self.raw_memories[i].number == number:
                return self.raw_memories[i]
        return None

    def get_raw_memory(self, number):
        return repr(self._get_raw_memory(number))

    @staticmethod
    def _decode_tone(toneval):
        # DCS examples:
        # D024N - 2814 - 0010 1000 0001 0100
        #                  ^--DCS
        # D024I - A814 - 1010 1000 0001 0100
        #                ^----inverted
        # D754I - A9EC - 1010 1001 1110 1100
        #    code in octal-------^^^^^^^^^^^

        pol = toneval & 0x8000 and 'R' or 'N'
        if toneval == 0xFFFF:
            return '', None, None
        elif toneval & 0x2000:
            # DTCS
            code = int('%o' % (toneval & 0x1FF))
            return 'DTCS', code, pol
        else:
            return 'Tone', toneval / 10.0, None

    @staticmethod
    def _encode_tone(mode, val, pol):
        if not mode:
            return 0xFFFF
        elif mode == 'Tone':
            return int(val * 10)
        elif mode == 'DTCS':
            code = int('%i' % val, 8)
            code |= 0x2800
            if pol == 'R':
                code |= 0x8000
            return code
        else:
            raise errors.RadioError('Unsupported tone mode %r' % mode)

    def get_memory(self, number):
        mem = chirp_common.Memory()
        mem.number = number
        _mem = self._get_raw_memory(number)
        if _mem is None:
            mem.empty = True
            return mem

        mem.name = str(_mem.name).rstrip('\x00')
        mem.freq = int(_mem.rx_freq) * 10
        chirp_common.split_tone_decode(mem,
                                       self._decode_tone(_mem.tx_tone),
                                       self._decode_tone(_mem.rx_tone))
        if _mem.wide:
            mem.mode = 'FM'
        else:
            mem.mode = 'NFM'

        mem.power = POWER_LEVELS[_mem.highpower]

        offset = (int(_mem.tx_freq) - int(_mem.rx_freq)) * 10
        if _mem.tx_freq.get_raw() == b'\xFF\xFF\xFF\xFF':
            mem.offset = 0
            mem.duplex = 'off'
        elif offset == 0:
            mem.duplex = ''
            mem.offset = 0
        elif abs(offset) < 10000000:
            mem.duplex = offset < 0 and '-' or '+'
            mem.offset = abs(offset)
        else:
            mem.duplex = 'split'
            mem.offset = int(_mem.tx_freq) * 10

        skipbyte = self._memobj.skipflags[(mem.number - 1) // 8]
        skipbit = skipbyte & (1 << (mem.number - 1) % 8)
        mem.skip = skipbit and 'S' or ''

        mem.extra = RadioSettingGroup('extra', 'Extra')
        mem.extra.append(MemSetting(
            'pttid', 'PTTID',
            RadioSettingValueMap(PTTID_SETTINGS.items(),
                                 int(_mem.pttid))))

        return mem

    def set_memory(self, mem):
        _mem = self._get_raw_memory(mem.number)
        if _mem is None:
            LOG.debug('Need to expand zone %i' % self._zone)

            # Calculate the new zone sizes and remap memory
            new_zones = [x[1] for x in self._parent._zones]
            new_zones[self._zone] = new_zones[self._zone] + 1
            self._parent.expand_mmap(new_zones)

            # Assign the new memory (at the end) to the desired
            # number
            _mem = self.raw_memories[self.raw_zoneinfo.count - 1]
            _mem.number = mem.number

            # Sort the memory into place
            self.shuffle_zone()

            # Now find it in the right spot
            _mem = self._get_raw_memory(mem.number)
            if _mem is None:
                raise errors.RadioError('Internal error after '
                                        'memory allocation')

            # Default values for unknown things
            _mem.unknown11 = 0x3
            _mem.unknown12 = 0x3
            _mem.unknown1 = [0xFF, 0xFF]
            _mem.unknown2 = [0xFF for i in range(0, 16)]
            _mem.unknown3_1 = 0xF
            _mem.unknown3_2 = 0x1
            _mem.unknown3_3 = 0x0
            _mem.unknown4 = 0xFF

            # Defaults for known things that might not always get set
            _mem.pttid = 0xFF

        if mem.empty:
            LOG.debug('Need to shrink zone %i' % self._zone)
            # Make the memory sort to the end, and sort the zone
            _mem.number = 0xFF
            self.shuffle_zone()

            # Calculate the new zone sizes and remap memory
            new_zones = [x[1] for x in self._parent._zones]
            new_zones[self._zone] = new_zones[self._zone] - 1
            self._parent.expand_mmap(new_zones)
            return

        _mem.name = mem.name[:12].encode().rstrip().ljust(12, b'\x00')
        _mem.rx_freq = mem.freq // 10

        txtone, rxtone = chirp_common.split_tone_encode(mem)
        _mem.tx_tone = self._encode_tone(*txtone)
        _mem.rx_tone = self._encode_tone(*rxtone)

        _mem.wide = mem.mode == 'FM'
        _mem.highpower = mem.power == POWER_LEVELS[1]

        if mem.duplex == '':
            _mem.tx_freq = mem.freq // 10
        elif mem.duplex == 'split':
            _mem.tx_freq = mem.offset // 10
        elif mem.duplex == 'off':
            _mem.tx_freq.set_raw(b'\xff\xff\xff\xff')
        elif mem.duplex == '-':
            _mem.tx_freq = (mem.freq - mem.offset) // 10
        elif mem.duplex == '+':
            _mem.tx_freq = (mem.freq + mem.offset) // 10
        else:
            raise errors.RadioError('Unsupported duplex mode %r' % mem.duplex)

        step_lookup = {
            12.5: 0x6,
            10.0: 0x5,
            6.25: 0x3,
            5: 0x2,
            2.5: 0x1,
        }
        _mem.rx_step = tk280.choose_step(step_lookup, int(_mem.rx_freq) * 10)
        _mem.tx_step = tk280.choose_step(step_lookup, int(_mem.tx_freq) * 10)

        skipbyte = self._memobj.skipflags[(mem.number - 1) // 8]
        if mem.skip == 'S':
            skipbyte |= (1 << (mem.number - 1) % 8)
        else:
            skipbyte &= ~(1 << (mem.number - 1) % 8)

        for setting in mem.extra:
            setting.apply_to_memobj(_mem)

    def _pure_choice_setting(self, settings_key, name, choices, default='Off'):
        if default is not None:
            ui_choices = [default] + choices
        else:
            ui_choices = choices
        s = RadioSetting(
            settings_key, name,
            RadioSettingValueList(
                ui_choices,
                get_choice(self._memobj.settings, settings_key,
                           choices, default)))
        s.set_apply_callback(set_choice, self._memobj.settings,
                             settings_key, choices, default)
        return s

    def _inverted_flag_setting(self, key, name, obj=None):
        if obj is None:
            obj = self._memobj.settings

        def apply_inverted(setting, key):
            setattr(obj, key, not int(setting.value))

        v = not getattr(obj, key)
        s = RadioSetting(
                key, name,
                RadioSettingValueBoolean(v))
        s.set_apply_callback(apply_inverted, key)
        return s

    def _get_common1(self):
        settings = self._memobj.settings
        common1 = RadioSettingSubGroup('common1', 'Common 1')

        common1.append(self._pure_choice_setting('sublcd',
                                                 'Sub LCD Display',
                                                 SUBLCD,
                                                 default='None'))

        def apply_clockfmt(setting):
            settings.clockfmt = CLOCKFMT.index(str(setting.value))

        clockfmt = RadioSetting(
            'clockfmt', 'Clock Format',
            RadioSettingValueList(CLOCKFMT,
                                  current_index=settings.clockfmt))
        clockfmt.set_apply_callback(apply_clockfmt)
        common1.append(clockfmt)

        def apply_datefmt(setting):
            settings.datefmt = DATEFMT.index(str(setting.value))

        datefmt = RadioSetting(
            'datefmt', 'Date Format',
            RadioSettingValueList(DATEFMT,
                                  current_index=settings.datefmt))
        datefmt.set_apply_callback(apply_datefmt)
        common1.append(datefmt)

        common1.append(self._pure_choice_setting('micsense',
                                                 'Mic Sense High',
                                                 MICSENSE))

        def apply_modereset(setting):
            val = int(setting.value)
            if val == 0:
                val = 0xFFFF
            settings.modereset = val

        _modereset = int(settings.modereset)
        if _modereset == 0xFFFF:
            _modereset = 0
        modereset = RadioSetting(
            'modereset', 'Mode Reset Timer',
            RadioSettingValueInteger(0, 300, _modereset))
        modereset.set_apply_callback(apply_modereset)
        common1.append(modereset)

        inverted_flags = [('power_switch_memory', 'Power Switch Memory'),
                          ('scrambler_memory', 'Scrambler Memory'),
                          ('off_hook_decode', 'Off-Hook Decode'),
                          ('ssi', 'Signal Strength Indicator'),
                          ('ignition_sense', 'Ingnition Sense')]
        for key, name in inverted_flags:
            if self.is_portable and key in ONLY_MOBILE_SETTINGS:
                # Skip settings that are not valid for portables
                continue
            common1.append(self._inverted_flag_setting(key, name))

        if not self.is_portable and 'ignition_mode' in ONLY_MOBILE_SETTINGS:
            common1.append(self._pure_choice_setting('ignition_mode',
                                                     'Ignition Mode',
                                                     ['Ignition & SW',
                                                      'Ignition Only'],
                                                     None))

        def apply_it(setting):
            settings.ignition_time = int(setting.value) / 600

        _it = int(settings.ignition_time) * 600
        it = RadioSetting(
            'it', 'Ignition Timer (s)',
            RadioSettingValueInteger(10, 28800, _it))
        it.set_apply_callback(apply_it)
        if not self.is_portable and 'it' in ONLY_MOBILE_SETTINGS:
            common1.append(it)

        return common1

    def _get_common2(self):
        settings = self._memobj.settings
        common2 = RadioSettingSubGroup('common2', 'Common 2')

        def apply_ponmsgtext(setting):
            settings.pon_msgtext = (
                str(setting.value)[:12].strip().ljust(12, '\x00'))

        common2.append(
            self._pure_choice_setting('pon_msgtype', 'Power On Message Type',
                                      STARTUP_MODES))

        _text = str(settings.pon_msgtext).rstrip('\x00')
        text = RadioSetting('settings.pon_msgtext',
                            'Power On Text',
                            RadioSettingValueString(
                                0, 12, _text))
        text.set_apply_callback(apply_ponmsgtext)
        common2.append(text)

        def apply_volume(setting, key):
            setattr(settings, key, VOLUMES[str(setting.value)])

        volumes = {'poweron_tone': 'Power-on Tone',
                   'control_tone': 'Control Tone',
                   'warning_tone': 'Warning Tone',
                   'alert_tone': 'Alert Tone',
                   'sidetone': 'Sidetone',
                   'locator_tone': 'Locator Tone'}
        for value, name in volumes.items():
            setting = getattr(settings, value)
            volume = RadioSetting('settings.%s' % value, name,
                                  RadioSettingValueList(
                                      VOLUMES.keys(),
                                      VOLUMES_REV.get(int(setting), 0)))
            volume.set_apply_callback(apply_volume, value)
            common2.append(volume)

        def apply_vol_level(setting, key):
            setattr(settings, key, int(setting.value))

        levels = {'lo_volume': 'Low Volume Level (Fixed Volume)',
                  'hi_volume': 'High Volume Level (Fixed Volume)',
                  'min_volume': 'Minimum Audio Volume',
                  'max_volume': 'Maximum Audio Volume'}
        for value, name in levels.items():
            setting = getattr(settings, value)
            if 'Audio' in name:
                minimum = 0
            else:
                minimum = 1
            volume = RadioSetting(
                'settings.%s' % value, name,
                RadioSettingValueInteger(minimum, 31, int(setting)))
            volume.set_apply_callback(apply_vol_level, value)
            common2.append(volume)

        def apply_vo(setting):
            val = int(setting.value)
            if val < 0:
                val = abs(val) | 0x80
            settings.tone_volume_offset = val

        _voloffset = int(settings.tone_volume_offset)
        if _voloffset & 0x80:
            _voloffset = abs(_voloffset & 0x7F) * -1
        voloffset = RadioSetting(
            'tvo', 'Tone Volume Offset',
            RadioSettingValueInteger(
                -5, 5,
                _voloffset))
        voloffset.set_apply_callback(apply_vo)
        common2.append(voloffset)

        def apply_mvp(setting):
            settings.min_vol_preset = MIN_VOL_PRESET[str(setting.value)]

        _volpreset = int(settings.min_vol_preset)
        volpreset = RadioSetting(
            'mvp', 'Minimum Volume Type',
            RadioSettingValueList(MIN_VOL_PRESET.keys(),
                                  MIN_VOL_PRESET_REV[_volpreset]))
        volpreset.set_apply_callback(apply_mvp)
        if not self.is_portable and 'mvp' in ONLY_MOBILE_SETTINGS:
            common2.append(volpreset)

        return common2

    def _get_common3(self):
        settings = self._memobj.settings
        group = RadioSettingSubGroup('common3', 'Common 3')

        rs = MemSetting('settings.battsave', 'Battery Save',
                        RadioSettingValueMap(BATTSAVE_SETTINGS.items(),
                                             settings.battsave))
        group.append(rs)

        rs = MemSetting('settings.battwarn', 'Battery Warning',
                        RadioSettingValueMap(BATTWARN_SETTINGS.items(),
                                             settings.battwarn))
        group.append(rs)

        rs = MemSetting('settings.battstatus', 'Battery Status',
                        RadioSettingValueInvertedBoolean(
                            not settings.battstatus))
        group.append(rs)

        rs = MemSetting('settings.pttid_type', 'PTTID Type',
                        RadioSettingValueMap(PTTID_TYPES.items(),
                                             settings.pttid_type))
        group.append(rs)

        def apply_dtmf(setting, which):
            setattr(settings, which, encode_dtmf(str(setting.value).strip()))

        code = decode_dtmf(settings.pttid_dtmf_bot_code)
        rs = RadioSetting('pttid_bot', 'Beginning of Transmit',
                          RadioSettingValueString(0, 16, code))
        rs.set_apply_callback(apply_dtmf, 'pttid_dtmf_bot_code')
        group.append(rs)

        code = decode_dtmf(settings.pttid_dtmf_eot_code)
        rs = RadioSetting('pttid_eot', 'End of Transmit',
                          RadioSettingValueString(0, 16, code))
        rs.set_apply_callback(apply_dtmf, 'pttid_dtmf_eot_code')
        group.append(rs)

        return group

    def _get_conventional(self):
        settings = self._memobj.settings

        conv = RadioSettingGroup('conv', 'Conventional')
        inverted_flags = [('busy_led', 'Busy LED'),
                          ('ost_memory', 'OST Status Memory'),
                          ('tone_off', 'Tone Off'),
                          ('ptt_release', 'PTT Release tone'),
                          ('ptt_proceed', 'PTT Proceed Tone')]
        for key, name in inverted_flags:
            conv.append(self._inverted_flag_setting(key, name))

        def apply_pttt(setting):
            settings.ptt_timer = int(setting.value)

        pttt = RadioSetting(
            'pttt', 'PTT Proceed Tone Timer (ms)',
            RadioSettingValueInteger(0, 6000, int(settings.ptt_timer)))
        pttt.set_apply_callback(apply_pttt)
        conv.append(pttt)

        self._get_ost(conv)

        return conv

    def _get_zones(self):
        zones = RadioSettingGroup('zones', 'Zones')

        zone_count = RadioSetting('_zonecount',
                                  'Number of Zones',
                                  RadioSettingValueInteger(
                                      1, 128, len(self._zones)))
        zone_count.set_doc('Number of zones in the radio. '
                           'Reducing this number '
                           'will DELETE memories in affected zones!')
        zone_count.set_volatile(True)
        zones.append(zone_count)

        for i in range(len(self._zones)):
            zone = RadioSettingSubGroup('zone%i' % i, 'Zone %i' % (i + 1))

            _zone = getattr(self._memobj, 'zone%i' % i).zoneinfo
            _name = str(_zone.name).rstrip('\x00')
            name = MemSetting('zone%i.zoneinfo.name' % i, 'Name',
                              RadioSettingValueString(0, 12, _name,
                                                      mem_pad_char='\x00'))
            zone.append(name)

            def apply_timer(setting, key, zone_number):
                val = int(setting.value)
                if val == 0:
                    val = 0xFFFF
                _zone = getattr(self._memobj, 'zone%i' % zone_number).zoneinfo
                setattr(_zone, key, val)

            def collapse(val):
                val = int(val)
                if val == 0xFFFF:
                    val = 0
                return val

            timer = RadioSetting(
                'z%itimeout' % i, 'Time-out Timer',
                RadioSettingValueInteger(15, 1200, collapse(_zone.timeout)))
            timer.set_apply_callback(apply_timer, 'timeout', i)
            zone.append(timer)

            timer = RadioSetting(
                'z%itot_alert' % i, 'TOT Pre-Alert',
                RadioSettingValueInteger(0, 10, collapse(_zone.tot_alert)))
            timer.set_apply_callback(apply_timer, 'tot_alert', i)
            zone.append(timer)

            timer = RadioSetting(
                'z%itot_rekey' % i, 'TOT Re-Key Time',
                RadioSettingValueInteger(0, 60, collapse(_zone.tot_rekey)))
            timer.set_apply_callback(apply_timer, 'tot_rekey', i)
            zone.append(timer)

            timer = RadioSetting(
                'z%itot_reset' % i, 'TOT Reset Time',
                RadioSettingValueInteger(0, 15, collapse(_zone.tot_reset)))
            timer.set_apply_callback(apply_timer, 'tot_reset', i)
            zone.append(timer)

            rs = MemSetting('zone%i.zoneinfo.bcl_override' % i,
                            'BCL Override',
                            RadioSettingValueInvertedBoolean(
                                not _zone.bcl_override))
            zone.append(rs)

            zones.append(zone)

        return zones

    def _get_ost(self, parent):
        tones = chirp_common.TONES[:]

        def apply_tone(setting, index, which):
            if str(setting.value) == 'Off':
                val = 0xFFFF
            else:
                val = int(float(str(setting.value)) * 10)
            setattr(self._memobj.ost_tones[index], '%stone' % which, val)

        def _tones():
            return ['Off'] + [str(x) for x in tones]

        ostgroup = RadioSettingGroup('ost', 'OST')
        ostgroup.set_doc('Operator Selectable Tone')
        parent.append(ostgroup)

        for i in range(0, 40):
            _ost = self._memobj.ost_tones[i]
            ost = RadioSettingSubGroup('ost%i' % i,
                                       'OST %i' % (i + 1))

            cur = str(_ost.name).rstrip('\x00')
            name = MemSetting('ost_tones[%i].name' % i, 'Name',
                              RadioSettingValueString(0, 12, cur,
                                                      mem_pad_char='\x00'))
            ost.append(name)

            if _ost.rxtone == 0xFFFF:
                cur = 'Off'
            else:
                cur = round(int(_ost.rxtone) / 10.0, 1)
                if cur not in tones:
                    LOG.debug('Non-standard OST rx tone %i %s' % (i, cur))
                    tones.append(cur)
                    tones.sort()
            rx = RadioSetting('rxtone%i' % i, 'RX Tone',
                              RadioSettingValueList(_tones(),
                                                    str(cur)))
            rx.set_apply_callback(apply_tone, i, 'rx')
            ost.append(rx)

            if _ost.txtone == 0xFFFF:
                cur = 'Off'
            else:
                cur = round(int(_ost.txtone) / 10.0, 1)
                if cur not in tones:
                    LOG.debug('Non-standard OST tx tone %i %s' % (i, cur))
                    tones.append(cur)
                    tones.sort()
            tx = RadioSetting('txtone%i' % i, 'TX Tone',
                              RadioSettingValueList(_tones(),
                                                    str(cur)))
            tx.set_apply_callback(apply_tone, i, 'tx')
            ost.append(tx)

            ostgroup.append(ost)

    def _get_keys(self):
        if self.is_mobile:
            model_functions = MOBILE_BUTTON_FUNCTIONS
        else:
            model_functions = PORTABLE_BUTTON_FUNCTIONS

        pri_functions = BUTTON_FUNCTIONS | model_functions
        sec_functions = (
            {k: v for k, v in (model_functions | BUTTON_FUNCTIONS).items()
             if v not in PRIMARY_ONLY})

        group = RadioSettingGroup('keys', 'Keys')
        pri = RadioSettingSubGroup('primary', 'Primary')
        sec = RadioSettingSubGroup('secondary', 'Secondary')
        buttons = self._memobj.buttons

        rs = MemSetting('selector_knob', 'Selector Knob',
                        RadioSettingValueMap(KNOB_MODE.items(),
                                             self._memobj.selector_knob))
        group.append(rs)

        rs = MemSetting('keypad_type', 'Keypad Type',
                        RadioSettingValueMap(KEYPAD_TYPE.items(),
                                             self._memobj.keypad_type))
        group.append(rs)

        rs = MemSetting('keypad_op', 'Keypad Operation',
                        RadioSettingValueMap(KEYPAD_OP.items(),
                                             self._memobj.keypad_op))
        group.append(rs)

        rs = MemSetting('list_selector_key', 'List Selector Key',
                        RadioSettingValueInvertedBoolean(
                            not self._memobj.list_selector_key))
        group.append(rs)

        for key, index in BUTTONS.items():
            rs = MemSetting('buttons[%i].primary' % index, key,
                            RadioSettingValueMap(pri_functions.items(),
                                                 buttons[index].primary))
            pri.append(rs)

            rs = MemSetting('buttons[%i].secondary' % index, key,
                            RadioSettingValueMap(sec_functions.items(),
                                                 buttons[index].secondary))
            sec.append(rs)
        group.append(pri)
        group.append(sec)

        return group

    def get_settings(self):
        zones = self._get_zones()
        common1 = self._get_common1()
        common2 = self._get_common2()
        common3 = self._get_common3()
        common = RadioSettingGroup('common', 'Common',
                                   common1, common2, common3)
        conv = self._get_conventional()
        keys = self._get_keys()
        top = RadioSettings(zones, common, conv, keys)
        return top

    def set_settings(self, settings):
        for element in settings.apply_to(self._memobj):
            if element.get_name() == '_zonecount':
                new_zone_count = int(element.value)
                zone_sizes = [x[1] for x in self._zones[:new_zone_count]]
                if len(self._zones) > new_zone_count:
                    self.expand_mmap(zone_sizes[:new_zone_count])
                elif len(self._zones) < new_zone_count:
                    self.expand_mmap(zone_sizes + (
                        [0] * (new_zone_count - len(self._zones))))
            elif element.has_apply_callback():
                element.run_apply_callback()

    def get_sub_devices(self):
        zones = []
        to_copy = ('VENDOR', 'MODEL', 'VALID_BANDS', '_model')
        for i, _ in enumerate(self._zones):
            zone = getattr(self._memobj, 'zone%i' % i)

            zone_cls = tk280.TKx80SubdevMeta.make_subdev(
                self, KenwoodTKx180RadioZone, zone, to_copy,
                VARIANT='Zone %s' % (
                    str(zone.zoneinfo.name).rstrip('\x00').rstrip()))

            zones.append(zone_cls(self, i))
        return zones


class KenwoodTKx180RadioZone:
    _zone = None

    def __init__(self, parent, zone=0):
        if isinstance(parent, KenwoodTKx180Radio):
            self._parent = parent
        else:
            LOG.warning('Parent was not actually our parent, expect failure')
        self._zone = zone

    @property
    def _zones(self):
        return self._parent._zones

    @property
    def _memobj(self):
        return self._parent._memobj

    def load_mmap(self, filename):
        self._parent.load_mmap(filename)

    def get_features(self):
        rf = KenwoodTKx180Radio.get_features(self)
        rf.has_sub_devices = False
        rf.memory_bounds = (1, 250)
        return rf

    def get_sub_devices(self):
        return []


@directory.register
class KenwoodTK7180Radio(KenwoodTKx180Radio):
    MODEL = 'TK-7180'
    VALID_BANDS = [(136000000, 174000000)]
    _model = b'M7180\x04'


@directory.register
class KenwoodTK8180Radio(KenwoodTKx180Radio):
    MODEL = 'TK-8180'
    VALID_BANDS = [(400000000, 520000000)]
    _model = b'M8180\x06'


@directory.register
class KenwoodTK2180Radio(KenwoodTKx180Radio):
    MODEL = 'TK-2180'
    VALID_BANDS = [(136000000, 174000000)]
    _model = b'P2180\x04'


# K1,K3 are technically 450-470 (K3 == keypad)
@directory.register
class KenwoodTK3180K1Radio(KenwoodTKx180Radio):
    MODEL = 'TK-3180K'
    VALID_BANDS = [(400000000, 520000000)]
    _model = b'P3180\x06'


# K2,K4 are technically 400-470 (K4 == keypad)
@directory.register
class KenwoodTK3180K2Radio(KenwoodTKx180Radio):
    MODEL = 'TK-3180K2'
    VALID_BANDS = [(400000000, 520000000)]
    _model = b'P3180\x07'


@directory.register
class KenwoodTK8180E(KenwoodTKx180Radio):
    MODEL = 'TK-8180E'
    VALID_BANDS = [(400000000, 520000000)]
    _model = b'M8189\''


@directory.register
class KenwoodTK7180ERadio(KenwoodTKx180Radio):
    MODEL = 'TK-7180E'
    VALID_BANDS = [(136000000, 174000000)]
    _model = b'M7189$'



# Copyright 2013 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from builtins import bytes

import unittest
from unittest import mock


from chirp import bitwise
from chirp import memmap


class BaseTest(unittest.TestCase):
    def _compare_structure(self, obj, primitive):
        for key, value in primitive.items():
            if isinstance(value, dict):
                self._compare_structure(getattr(obj, key), value)
            else:
                self.assertEqual(type(value)(getattr(obj, key)), value)


class TestMemoryMapCoherence(BaseTest):
    def test_byte_char_coherence(self):
        charmmap = memmap.MemoryMap('00')
        # This will to a get_byte_compatible() from chars
        obj = bitwise.parse('char foo[2];', charmmap)
        self.assertEqual('00', str(obj.foo))
        obj.foo = '11'
        # The above assignment happens on the byte-compatible mmap,
        # make sure it is still visible in the charmmap we know about.
        # This confirms that get_byte_compatible() links the backing
        # store of the original mmap to the new one.
        self.assertEqual('11', charmmap.get_packed())


class TestBitwiseBaseIntTypes(BaseTest):
    def _test_type(self, datatype, _data, value):
        data = memmap.MemoryMapBytes(bytes(_data))
        obj = bitwise.parse("%s foo;" % datatype, data)
        self.assertEqual(int(obj.foo), value)
        self.assertEqual(obj.foo.size(), len(data) * 8)

        obj.foo = 0
        self.assertEqual(int(obj.foo), 0)
        self.assertEqual(data.get_packed(), (b"\x00" * (obj.size() // 8)))

        obj.foo = value
        self.assertEqual(int(obj.foo), value)
        self.assertEqual(data.get_packed(), _data)

        obj.foo = 7
        # Compare against the equivalent real division so we get consistent
        # results on py2 and py3
        self.assertEqual(7 // 2, obj.foo // 2)
        self.assertEqual(7 / 2, obj.foo / 2)
        self.assertEqual(7 / 2.0, obj.foo / 2.0)

    def test_type_u8(self):
        self._test_type("u8", b"\x80", 128)

    def test_type_u16(self):
        self._test_type("u16", b"\x01\x00", 256)

    def test_type_u24(self):
        self._test_type("u24", b"\x80\x00\x00", 2**23)

    def test_type_u32(self):
        self._test_type("u32", b"\x80\x00\x00\x00", 2**31)

    def test_type_ul16(self):
        self._test_type("ul16", b"\x00\x01", 256)

    def test_type_ul24(self):
        self._test_type("ul24", b"\x00\x00\x80", 2**23)

    def test_type_ul32(self):
        self._test_type("ul32", b"\x00\x00\x00\x80", 2**31)

    def test_int_array(self):
        data = memmap.MemoryMapBytes(bytes(b'\x00\x01\x02\x03'))
        obj = bitwise.parse('u8 foo[4];', data)
        for i in range(4):
            self.assertEqual(i, obj.foo[i])
            obj.foo[i] = i * 2
        self.assertEqual(b'\x00\x02\x04\x06', data.get_packed())

    def test_int_array_of_one(self):
        data = memmap.MemoryMapBytes(bytes(b'\x00\x01\x02\x03'))
        obj = bitwise.parse('u8 foo[1];', data)
        self.assertEqual(0, obj.foo[0])
        obj.foo[0] = 1
        self.assertEqual(b'\x01\x01\x02\x03', data.get_packed())

    def test_int_array_of_zero(self):
        data = memmap.MemoryMapBytes(bytes(b'\x00\x01\x02\x03'))
        obj = bitwise.parse('u8 foo[0x0]; u8 bar;', data)
        # Make sure we can't access any elements of zero-length foo
        self.assertRaises(IndexError, lambda: obj.foo[0])
        # Make sure bar is still at the front, unaffected by zero-length foo
        self.assertEqual(0, obj.bar)

    def test_int_array_set_raw(self):
        data = memmap.MemoryMapBytes(bytes(b'\x00\x01\x02\x03'))
        obj = bitwise.parse('u8 foo[4];', data)
        obj.foo.set_raw(b'\x09\x08\x07\x06')
        self.assertEqual(9, obj.foo[0])
        self.assertEqual(8, obj.foo[1])
        self.assertEqual(7, obj.foo[2])
        self.assertEqual(6, obj.foo[3])

        obj = bitwise.parse('u16 foo[2];', data)
        obj.foo.set_raw(b'\x00\x01\x00\x02')
        self.assertEqual(1, obj.foo[0])
        self.assertEqual(2, obj.foo[1])

        with self.assertRaises(AssertionError):
            obj.foo.set_raw(b'123')

        with self.assertRaises(AssertionError):
            obj.foo.set_raw(b'12345')


class TestBitfieldTypes(BaseTest):
    def test_bitfield_u8(self):
        defn = "u8 foo:4, bar:4;"
        data = memmap.MemoryMapBytes(bytes(b"\x12"))
        obj = bitwise.parse(defn, data)
        self.assertEqual(obj.foo, 1)
        self.assertEqual(obj.bar, 2)
        self.assertEqual(obj.foo.size(), 4)
        self.assertEqual(obj.bar.size(), 4)
        obj.foo = 0x8
        obj.bar = 0x1
        self.assertEqual(data.get_packed(), b"\x81")

    def _test_bitfield_16(self, variant, data):
        defn = "u%s16 foo:4, bar:8, baz:4;" % variant
        data = memmap.MemoryMapBytes(bytes(data))
        obj = bitwise.parse(defn, data)
        self.assertEqual(int(obj.foo), 1)
        self.assertEqual(int(obj.bar), 0x23)
        self.assertEqual(int(obj.baz), 4)
        self.assertEqual(obj.foo.size(), 4)
        self.assertEqual(obj.bar.size(), 8)
        self.assertEqual(obj.baz.size(), 4)
        obj.foo = 0x2
        obj.bar = 0x11
        obj.baz = 0x3
        if variant == "l":
            self.assertEqual(data.get_packed(), b"\x13\x21")
        else:
            self.assertEqual(data.get_packed(), b"\x21\x13")

    def test_bitfield_u16(self):
        self._test_bitfield_16("", b"\x12\x34")

    def test_bitfield_ul16(self):
        self._test_bitfield_16('l', b"\x34\x12")

    def _test_bitfield_24(self, variant, data):
        defn = "u%s24 foo:12, bar:6, baz:6;" % variant
        data = memmap.MemoryMapBytes(bytes(data))
        obj = bitwise.parse(defn, data)
        self.assertEqual(int(obj.foo), 4)
        self.assertEqual(int(obj.bar), 3)
        self.assertEqual(int(obj.baz), 2)
        self.assertEqual(obj.foo.size(), 12)
        self.assertEqual(obj.bar.size(), 6)
        self.assertEqual(obj.baz.size(), 6)
        obj.foo = 1
        obj.bar = 2
        obj.baz = 3
        if variant == 'l':
            self.assertEqual(data.get_packed(), b"\x83\x10\x00")
        else:
            self.assertEqual(data.get_packed(), b"\x00\x10\x83")

    def test_bitfield_u24(self):
        self._test_bitfield_24("", b"\x00\x40\xC2")

    def test_bitfield_ul24(self):
        self._test_bitfield_24("l", b"\xC2\x40\x00")


class TestBitType(BaseTest):
    def test_bit_array(self):
        defn = "bit foo[24];"
        data = memmap.MemoryMapBytes(bytes(b"\x00\x80\x01"))
        obj = bitwise.parse(defn, data)
        for i, v in [(0, False), (8, True), (23, True)]:
            self.assertEqual(bool(obj.foo[i]), v)
        for i in range(0, 24):
            obj.foo[i] = i % 2
        self.assertEqual(data.get_packed(), b"\x55\x55\x55")

    def test_lbit_array(self):
        defn = "lbit foo[24];"
        data = memmap.MemoryMapBytes(bytes(b"\x00\x20\x80"))
        obj = bitwise.parse(defn, data)
        for i, v in [(0, False), (13, True), (23, True)]:
            self.assertEqual(bool(obj.foo[i]), v)
        for i in range(0, 24):
            obj.foo[i] = i % 2
        self.assertEqual(data.get_packed(), b"\xAA\xAA\xAA")

    def test_bit_array_fail(self):
        self.assertRaises(ValueError, bitwise.parse, "bit foo[23];", b"000")


class TestBitwiseBCDTypes(BaseTest):
    def _test_def(self, definition, name, _data, value):
        data = memmap.MemoryMapBytes(bytes(_data))
        obj = bitwise.parse(definition, data)
        self.assertEqual(int(getattr(obj, name)), value)
        self.assertEqual(getattr(obj, name).size(), len(_data) * 8)
        setattr(obj, name, 0)
        self.assertEqual(data.get_packed(), (b"\x00" * len(_data)))
        setattr(obj, name, 42)
        if definition.startswith("b"):
            expected = (len(_data) == 2 and b"\x00" or b"") + b"\x42"
        else:
            expected = b"\x42" + (len(_data) == 2 and b"\x00" or b"")
        raw = data.get_packed()
        self.assertEqual(raw, expected)

    def test_bbcd(self):
        self._test_def("bbcd foo;", "foo", b"\x12", 12)

    def test_lbcd(self):
        self._test_def("lbcd foo;", "foo", b"\x12", 12)

    def test_bbcd_array(self):
        self._test_def("bbcd foo[2];", "foo", b"\x12\x34", 1234)

    def test_lbcd_array(self):
        self._test_def("lbcd foo[2];", "foo", b"\x12\x34", 3412)

    def test_bbcd_array_leading_bits(self):
        data = memmap.MemoryMapBytes(bytes(b"\xc7\x54"))
        obj = bitwise.parse("bbcd foo[2];", data)
        # This is what we interpret a leading 0xC000 as
        self.assertEqual(int(obj.foo), 754 + 12000)

        # Ignore the top nibble and we expect just the bottom three
        obj.foo[0].ignore_bits(0xF0)
        self.assertEqual(int(obj.foo), 754)

        # If wet try to set something in the ignored nibble it should be
        # ignored but preserved
        obj.foo = 1754
        self.assertEqual(int(obj.foo), 754)
        self.assertEqual(data[0][0], 0xC7)


class TestBitwiseCharTypes(BaseTest):
    def test_char(self):
        data = memmap.MemoryMapBytes(bytes(b"c"))
        obj = bitwise.parse("char foo;", data)
        self.assertEqual(str(obj.foo), "c")
        self.assertEqual(obj.foo.size(), 8)
        obj.foo = "d"
        self.assertEqual(data.get_packed(), b"d")

    def test_string(self):
        data = memmap.MemoryMapBytes(bytes(b"foobar"))
        obj = bitwise.parse("char foo[6];", data)
        self.assertEqual(str(obj.foo), "foobar")
        self.assertEqual(obj.foo.size(), 8 * 6)
        obj.foo = "bazfoo"
        self.assertEqual(data.get_packed(), b"bazfoo")

    def test_string_invalid_chars(self):
        data = memmap.MemoryMapBytes(bytes(b"\xFFoobar1"))
        obj = bitwise.parse("struct {char foo[7];} bar;", data)

        expected = '\xffoobar1'

        self.assertIn(expected, repr(obj.bar))

    def test_string_wrong_length(self):
        data = memmap.MemoryMapBytes(bytes(b"foobar"))
        obj = bitwise.parse("char foo[6];", data)
        self.assertRaises(ValueError, setattr, obj, "foo", "bazfo")
        self.assertRaises(ValueError, setattr, obj, "foo", "bazfooo")

    def test_string_with_various_input_types(self):
        data = memmap.MemoryMapBytes(bytes(b"foobar"))
        obj = bitwise.parse("char foo[6];", data)
        self.assertEqual('foobar', str(obj.foo))
        self.assertEqual(6, len(b'barfoo'))
        obj.foo = b'barfoo'
        self.assertEqual('barfoo', str(obj.foo))
        obj.foo = [ord(c) for c in 'fffbbb']
        self.assertEqual('fffbbb', str(obj.foo))

    def test_string_get_raw(self):
        data = memmap.MemoryMapBytes(bytes(b"foobar"))
        obj = bitwise.parse("char foo[6];", data)
        self.assertEqual(b'foobar', obj.foo.get_raw())
        self.assertEqual('foobar', obj.foo.get_raw(asbytes=False))


class TestBitwiseStructTypes(BaseTest):
    def _test_def(self, definition, data, primitive):
        obj = bitwise.parse(definition, data)
        self._compare_structure(obj, primitive)
        self.assertEqual(obj.size(), len(data) * 8)

    def test_struct_one_element(self):
        defn = "struct { u8 bar; } foo;"
        value = {"foo": {"bar": 128}}
        self._test_def(defn, b"\x80", value)

    def test_struct_two_elements(self):
        defn = "struct { u8 bar; u16 baz; } foo;"
        value = {"foo": {"bar": 128, "baz": 256}}
        self._test_def(defn, b"\x80\x01\x00", value)

    def test_struct_writes(self):
        data = memmap.MemoryMapBytes(bytes(b".."))
        defn = "struct { u8 bar; u8 baz; } foo;"
        obj = bitwise.parse(defn, data)
        obj.foo.bar = 0x12
        obj.foo.baz = 0x34
        self.assertEqual(data.get_packed(), b"\x12\x34")

    def test_struct_get_raw(self):
        data = memmap.MemoryMapBytes(bytes(b".."))
        defn = "struct { u8 bar; u8 baz; } foo;"
        obj = bitwise.parse(defn, data)
        self.assertEqual(b'..', obj.get_raw())
        self.assertEqual('..', obj.get_raw(asbytes=False))

    def test_struct_get_raw_small(self):
        data = memmap.MemoryMapBytes(bytes(b"."))
        defn = "struct { u8 bar; } foo;"
        obj = bitwise.parse(defn, data)
        self.assertEqual(b'.', obj.get_raw())
        self.assertEqual('.', obj.get_raw(asbytes=False))

    def test_struct_set_raw(self):
        data = memmap.MemoryMapBytes(bytes(b"."))
        defn = "struct { u8 bar; } foo;"
        obj = bitwise.parse(defn, data)
        obj.set_raw(b'1')
        self.assertEqual(b'1', data.get_packed())
        obj.set_raw('2')
        self.assertEqual(b'2', data.get_packed())

    @mock.patch.object(bitwise.LOG, 'error')
    def test_struct_duplicate(self, mock_log):
        bitwise.parse('struct\n{ u8 foo; u8 foo1:2, foo:4, foo3:2;} bar;',
                      memmap.MemoryMapBytes(b'\x00' * 128))
        bitwise.parse('struct\n{ u8 foo; u8 foo;} bar;',
                      memmap.MemoryMapBytes(b'\x00' * 128))
        bitwise.parse('struct\n{ u8 foo; u8 foo[2];} bar;',
                      memmap.MemoryMapBytes(b'\x00' * 128))
        self.assertEqual(3, mock_log.call_count)

    def test_struct_fill_raw(self):
        data = memmap.MemoryMapBytes(bytes(b"..."))
        defn = "struct { u8 bar; u16 baz; } foo;"
        obj = bitwise.parse(defn, data)
        obj.fill_raw(b'\xAA')
        self.assertEqual(0xAA, obj.foo.bar)
        self.assertEqual(0xAAAA, obj.foo.baz)
        obj.foo.fill_raw(b'\xBB')
        self.assertEqual(0xBB, obj.foo.bar)
        self.assertEqual(0xBBBB, obj.foo.baz)
        obj.foo.bar.fill_raw(b'\xCC')
        self.assertEqual(0xCC, obj.foo.bar)
        self.assertEqual(0xBBBB, obj.foo.baz)

        self.assertRaises(AssertionError, obj.fill_raw, '1')
        self.assertRaises(AssertionError, obj.fill_raw, b'AB')
        self.assertRaises(AssertionError, obj.fill_raw, False)

    def test_struct_array_of_one(self):
        data = memmap.MemoryMapBytes(bytes(b"123"))
        defn = "struct { u8 bar; } foo[1];"
        obj = bitwise.parse(defn, data)
        self.assertEqual(ord('1'), obj.foo[0].bar)


class TestBitwiseUnionTypes(BaseTest):
    def test_union_one_element(self):
        defn = 'union { u8 bar; } foo;'
        data = memmap.MemoryMapBytes(bytes(b'\x80'))
        obj = bitwise.parse(defn, data)
        self.assertEqual(128, obj.foo.bar)

    def test_union_two_elements(self):
        defn = 'union { u8 bar; u8 baz; } foo;'
        data = memmap.MemoryMapBytes(bytes(b'\x80'))
        obj = bitwise.parse(defn, data)
        self.assertEqual(128, obj.foo.bar)
        self.assertEqual(128, obj.foo.baz)
        obj.foo.baz = 1
        self.assertEqual(1, obj.foo.baz)
        self.assertEqual(1, obj.foo.bar)
        obj.foo.bar = 2
        self.assertEqual(2, obj.foo.baz)
        self.assertEqual(2, obj.foo.bar)

    def test_empty_union(self):
        defn = 'union { } foo;'
        data = memmap.MemoryMapBytes(bytes(b'\x80'))
        self.assertRaises(SyntaxError, bitwise.parse, defn, data)

    def test_union_size_mismatch(self):
        defn = 'union { u16 bar; u8 baz; } foo;'
        data = memmap.MemoryMapBytes(bytes(b'\x80'))
        self.assertRaises(bitwise.ParseError, bitwise.parse, defn, data)

    def test_union_adjacent(self):
        # Make sure that unions don't mess up adjacent fields
        defn = 'u8 pre; union { u8 bar; u8 baz; } foo; u8 post;'
        data = memmap.MemoryMapBytes(bytes(b'\x01\x02\x03'))
        obj = bitwise.parse(defn, data)
        self.assertEqual(1, obj.pre)
        self.assertEqual(2, obj.foo.bar)
        self.assertEqual(2, obj.foo.baz)
        self.assertEqual(3, obj.post)
        obj.pre = 4
        obj.foo.baz = 5
        obj.post = 6
        self.assertEqual(4, obj.pre)
        self.assertEqual(5, obj.foo.bar)
        self.assertEqual(5, obj.foo.baz)
        self.assertEqual(6, obj.post)

    def test_complex_union(self):
        defn = """u8 pre;
                  union {
                      u16 whole;
                      struct { u8 lo; u8 hi; } bytes;
                      union { u16 big; ul16 little; } endian;
                      bbcd digits[2];
                  } foo;
                  u8 post;
                  """
        data = memmap.MemoryMapBytes(bytes(b'\x01\x02\x03\x04'))
        obj = bitwise.parse(defn, data)

        self.assertEqual(1, obj.pre)
        self.assertEqual(4, obj.post)
        self.assertEqual(0x0203, obj.foo.whole)
        self.assertEqual(2, obj.foo.bytes.lo)
        self.assertEqual(3, obj.foo.bytes.hi)
        self.assertEqual(0x0203, obj.foo.endian.big)
        self.assertEqual(0x0302, obj.foo.endian.little)
        self.assertEqual(2, int(obj.foo.digits[0]))
        self.assertEqual(3, int(obj.foo.digits[1]))

        obj.foo.bytes.lo = 5
        self.assertEqual(1, obj.pre)
        self.assertEqual(4, obj.post)
        self.assertEqual(0x0503, obj.foo.whole)
        self.assertEqual(5, obj.foo.bytes.lo)
        self.assertEqual(3, obj.foo.bytes.hi)
        self.assertEqual(0x0503, obj.foo.endian.big)
        self.assertEqual(0x0305, obj.foo.endian.little)
        self.assertEqual(5, int(obj.foo.digits[0]))
        self.assertEqual(3, int(obj.foo.digits[1]))

    def test_union_array(self):
        defn = 'union { u8 bar; u8 baz; } foo[2];'
        data = memmap.MemoryMapBytes(bytes(b'\x01\x02'))
        obj = bitwise.parse(defn, data)
        self.assertEqual(1, obj.foo[0].bar)
        self.assertEqual(1, obj.foo[0].baz)
        self.assertEqual(2, obj.foo[1].bar)
        self.assertEqual(2, obj.foo[1].baz)
        obj.foo[0].bar = 3
        obj.foo[1].baz = 4
        self.assertEqual(3, obj.foo[0].bar)
        self.assertEqual(3, obj.foo[0].baz)
        self.assertEqual(4, obj.foo[1].bar)
        self.assertEqual(4, obj.foo[1].baz)

    def test_struct_size_with_union(self):
        # Make sure to include a bitfield at the start of the union to
        # confirm that we properly account for the union size separately.
        defn = ('struct { u8 pre; union { u8 bar1:4, bar2:4; u8 baz; } foo; '
                'u8 post; } st;')
        data = memmap.MemoryMapBytes(bytes(b'\x01\x02\x03'))
        obj = bitwise.parse(defn, data)
        # Make sure the union is counted as a single byte
        self.assertEqual(8 * 3, obj.st.size())

        # Make sure raw operations work across unions
        obj.st.set_raw(b'\x00' * 3)
        self.assertEqual(0, obj.st.foo.bar1)
        obj.st.fill_raw(b'\xFF')
        self.assertEqual(0x0F, obj.st.foo.bar1)
        self.assertEqual(b'\xFF\xFF\xFF', obj.st.get_raw())


class TestBitwisePrintoffset(BaseTest):
    @mock.patch.object(bitwise.LOG, 'debug')
    def test_printoffset(self, mock_log):
        defn = 'u8 foo; u16 bar; #printoffset "bar";'
        bitwise.parse(defn, b"abcdZ")
        mock_log.assert_called_once_with('bar: 3 (0x00000003)')


class TestBitwiseSeek(BaseTest):
    def test_seekto(self):
        defn = "#seekto 4; char foo;"
        obj = bitwise.parse(defn, b"abcdZ")
        self.assertEqual(str(obj.foo), "Z")

    def test_seek(self):
        defn = "char foo; #seek 3; char bar;"
        obj = bitwise.parse(defn, b"AbcdZ")
        self.assertEqual(str(obj.foo), "A")
        self.assertEqual(str(obj.bar), "Z")


class TestBitwiseErrors(BaseTest):
    def test_missing_semicolon(self):
        self.assertRaises(SyntaxError, bitwise.parse, "u8 foo", "")

    def test_missing_element(self):
        obj = bitwise.parse('u8 foo; struct { u8 bar; } baz;', b'12')
        self.assertRaisesRegex(AttributeError,
                               'No attribute does_not_exist',
                               getattr, obj, 'does_not_exist')
        self.assertRaisesRegex(AttributeError,
                               'No attribute this_either',
                               getattr, obj.baz, 'this_either')
        self.assertTrue('foo' in obj)
        self.assertFalse('does_not_exist' in obj)


class TestBitwiseComments(BaseTest):
    def test_comment_inline_cppstyle(self):
        obj = bitwise.parse('u8 foo; // test', b'\x10')
        self.assertEqual(16, obj.foo)

    def test_comment_cppstyle(self):
        obj = bitwise.parse('// Test this\nu8 foo;', b'\x10')
        self.assertEqual(16, obj.foo)


class TestBitwiseStringEncoders(BaseTest):
    def test_encode_bytes(self):
        self.assertEqual(b'foobar\x00',
                         bitwise.string_straight_encode('foobar\x00'))

    def test_decode_bytes(self):
        self.assertEqual('foobar\x00',
                         bitwise.string_straight_decode(b'foobar\x00'))


class TestPath(BaseTest):
    def test_get_path(self):
        fmt = ("u8 foo;"
               "u8 bar[2];"
               "struct {"
               "  u8 foo;"
               "  u8 bar[2];"
               "  u8 baz1:4,"
               "     baz2:4;"
               "  struct {"
               "    u16 childitem;"
               "  } child;"
               "} structure[2];")
        obj = bitwise.parse(fmt, memmap.MemoryMapBytes(b'\x00' * 128))
        obj.structure[0].bar[1] = 123
        obj.structure[1].child.childitem = 456
        self.assertEqual(123, obj.get_path('.structure[0].bar[1]'))
        self.assertEqual(456, obj.get_path('structure[1].child.childitem'))



# TK-690 support and improvements added in 2024 by Patrick Leiser,
# based on the work of:
# Copyright 2016-2024 Pavel Milanes CO7WT, <pavelmc@gmail.com>
#
# And with the help of Tom Hayward, who gently provided me with a driver he
# started and never finished for this radio.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import struct

from chirp import chirp_common, directory, memmap, errors, util, bitwise
from chirp.settings import RadioSettingGroup, RadioSetting, RadioSettings, \
    RadioSettingValueBoolean, RadioSettingValueList, RadioSettingValueString, \
    RadioSettingValueMap, RadioSettingValueInteger

LOG = logging.getLogger(__name__)

# Note: the exported .dat files from the official KPG-44D software have a
# metadata preable for the first 64 bytes, then contain the desired image
# to export to the radio itself (which is what chirp handles).

# IMPORTANT MEM DATA #########################################
# This radios have a odd mem structure, it seems like you have to
# manage 3 memory sectors that we concatenate on just one big
# memmap, as follow
#
# Low memory (Main CPU)
#    0x0000 to 0x4000
# Mid memory (Unknown)
#    0x4000 to 0x4090
# High memory (Head)
#    0x4090 to 0x6090
###############################################################

MEM_FORMAT = """

//# seekto 0x0000;
struct {
    u8 unknown[16];
    u8 ch_name_length;
    u8 grp_name_length;
} settings;

#seekto 0x0300;
struct {
    u8 grp_up;        // functions: 0x0 through 0x2 = Aux A through C
    u8 grp_down;      //          0x3 through 0x7 = Ch 1 through 5 direct,
    u8 monitor;       //          0x8 = Ch down, 0x9 = Ch up, 0xA = ch name,
    u8 scan;          //          0xB = Ch recall, 0xC = Del/Add, 0xD = dimmer,
    u8 PF1;           //          0xE = Emergency Call, 0xF = Grp down,
    u8 PF2;           //          0x10 = Grp up, 0x11 = HC1 (fixed),
    u8 PF3;           //          0x12 = HC2 (toggle), 0x13 = Horn Alert
    u8 PF4;           //          0x16 = Monitor, 0x17 = Operator Sel tone
    u8 PF5;           //          0x19 = Public Address, 0x1A = Scan,
    u8 PF6;           //          0x1C = Speaker int/ext, 0x1D = Squelch,
    u8 PF7;           //          0x1E= Talk Around, 0xFF = no function
    u8 PF8;
    u8 PF9;
    u8 unknown1;             // 0x10 when full head used
    u8 unknown2;             // 0x0F when full head used
    u8 unknown3[12];         //all 0xFF
    u8 knob_control;            //00 = ch up/down, 01 = grp up/down
    u8 headUnitType;            //0x00 = full, 0xFF = basic
} button_assignments;

#seekto 0x0400;
struct {
    u8 tot;                     // * 10 seconds, 30 sec increments
    u8 tot_pre_alert;           // seconds, off - 10, default 5
    u8 tot_rekey_time;          // seconds, off - 60, default off
    u8 tot_reset_time;          // seconds, off - 15, default off
    u8 unknown[4];
} group_settings[160];

#seekto 0x1480;
struct {
    u8 index;          // the index in the group_belong where this group start
    u8 length;         // how many channels are in this group
} group_limits[160];

#seekto 0x1600;
struct {
    u8 number;         // CH number relative to the group, 1-160
    u8 index;          // index relative to the memory space memory[index]
} group_belong[160];

#seekto 0x1800;
struct {
  lbcd rxfreq[4];       // 00-03
  lbcd txfreq[4];       // 04-07
  ul16 rxtone;
  ul16 txtone;
  u8 unknown0:1,
     power:1,            // power 0 = high, 1 = low
     beatshift:1,        // beat shift, 1 = on
     bcl:1,              // busy channel lockout, 1 = on
     pttid:1,            // ptt id, 1 = on
     signal:3;           //off=0, 1=DTMF, 2,3,4 = "2-Tone 1,2,3"
  u8 unknown1:2,
     txdisable:1,        // 1=RX only, 0=RX/TX
     unknown7:1,
     add:1,              // scan add, 1 = add
     wide:1,             // Wide = 1, narrow = 0
     unknown2:1,
     unknown4:1;
  u8 unknown5;
  u8 unknown6:6,
     compander:1,     // 0 = compander active, only applicable on narrow band
     valid:1;        // 0 = valid entry, enabled; 1=invalid
} memory[160];

#seekto 0x3DF0;
char poweron_msg[14];

#seekto 0x3E80;
struct {
  char line1[32];
  char line2[32];
} embeddedMessage;

#seekto 0x3ED0;
struct {
  u8 unknown10[10];
  char soft[6];
  struct {
    char variant[5];
    u8 unknown1;
    u8 memorysource;  //0x50 if exported from KPG44D, 0x00 if cloned from radio
    u8 unknown2:5,
       headtype:1,    //0 = basic head, 4 = full head
       unknown3:2;
    u8 unknown4;
    u8 unknown5;
  } rid;
  u8 unknown11[6];
  u8 unknown12[11];
  char soft_ver[5];
} properties;

struct {
  lbcd rxfreq[4];       // 00-03
  lbcd txfreq[4];       // 04-07
} test_frequencies[16];

#seekto 0x40A0;
struct {
  char name[16];
} grp_names[160];

struct {
  char name[16];
} chs_names[160];

"""

MEM_SIZE = 0x60A0  # 24,736 bytes
DAT_FILE_SIZE = 0x60E0
ACK_CMD = b'\x06'
RX_BLOCK_SIZE_L = 128
MEM_LR = range(0x0380, 0x0400)
RX_BLOCK_SIZE_M = 16
MEM_MR = range(1, 11)
RX_BLOCK_SIZE_H = 32
MEM_HR = range(0, 0x2000, RX_BLOCK_SIZE_H)
EMPTY_BLOCK = b"\xFF" * 256
EMPTY_L = b"\xFF" * RX_BLOCK_SIZE_L
EMPTY_H = b"\xFF" * RX_BLOCK_SIZE_H
VALID_CHARS = chirp_common.CHARSET_UPPER_NUMERIC + "()/\\*@-+,.#_"
VALID_CHARS_UNCASED = VALID_CHARS+VALID_CHARS.lower()
SKIP_VALUES = ["S", ""]
TONES = chirp_common.TONES
DTCS_CODES = chirp_common.DTCS_CODES
OPTSIG_LIST = ["None", "DTMF", "2-Tone 1", "2-Tone 2", "2-Tone 3"]
HEAD_TYPES = ["Basic", "Full"]

BUTTON_FUNCTION_LIST = [
            ('Aux A', 0), ('Aux B', 1), ('Aux C', 2),
            ('Ch 1 direct', 3), ('Ch 2 direct', 4), ('Ch 3 direct', 5),
            ('Ch 4 direct', 6), ('Ch 5 direct', 7), ('Ch down', 8),
            ('Ch up', 9), ('Ch name', 10), ('Ch recall', 11), ('Del/Add', 12),
            ('Dimmer', 13), ('Emergency Call', 14), ('Grp down', 15),
            ('Grp up', 16), ('HC1 (fixed)', 17), ('HC2 (toggle)', 18),
            ('Horn Alert', 19), ('Monitor', 22), ('Operator Sel tone', 23),
            ('Public Address', 25), ('Scan', 26), ('Speaker int/ext', 28),
            ('Squelch', 29), ('Talk Around', 30), ('no function', 255)]

ASSIGNABLE_BUTTONS = ["grp_up", "grp_down", "monitor", "scan", "PF1", "PF2",
                      "PF3", "PF4", "PF5", "PF6", "PF7", "PF8", "PF9"]
FULL_HEAD_ONLY_BUTTONS = ["monitor", "scan", "PF6", "PF7", "PF8", "PF9"]


def _close_radio(radio):
    """Get the radio out of program mode"""
    try:
        radio.pipe.write(b"E")
    except Exception:
        LOG.debug("Failed to close radio, serial error")
        raise errors.RadioError("Serial Connection Error while closing radio")


def _checksum(data):
    """the radio block checksum algorithm"""
    cs = 0
    for byte in data:
        cs += byte
    return cs % 256


def _make_framel(cmd, addr):
    """Pack the info in the format it likes"""
    # x52 x0F (x0380-x0400)
    return struct.pack(">BBH", ord(cmd), 0x0F, addr)


def _make_framem(cmd, addr):
    """Pack the info in the format it likes"""
    # x54 x0F (x00-x0A)
    return struct.pack(">BBB", ord(cmd), 0x0F, addr)


def _make_frameh(cmd, addr):
    """Pack the info in the format it likes"""
    # x53 x8F (x0000-x2000) x20
    return struct.pack(">BBHB", ord(cmd), 0x8F, addr, RX_BLOCK_SIZE_H)


def _handshake(radio, msg="", full=True):
    """Make a full handshake"""
    if full is True:
        radio.pipe.write(ACK_CMD)

    ack = radio.pipe.read(1)
    if ack != ACK_CMD:
        mesg = "Handshake failed, got ack: '0x%02x': %s" % (ord(ack), msg)
        LOG.debug(mesg)
        raise errors.RadioError("Radio failed to acknowledge our command")


def _recvl(radio):
    """Receive low data block from the radio, 130 or 2 bytes"""
    rxdata = radio.pipe.read(2)
    if rxdata == b"\x5A\xFF":
        # when the RX block has 2 bytes and the paylod+CS is \x5A\xFF
        # then the block is all \xFF
        _handshake(radio, "short block")
        return False
    rxdata += radio.pipe.read(RX_BLOCK_SIZE_L)
    if len(rxdata) == RX_BLOCK_SIZE_L + 2 and rxdata[0] == b"W"[0]:
        # Data block is W + Data(128) + CS
        rcs = rxdata[-1]
        data = rxdata[1:-1]
        ccs = _checksum(data)

        if rcs != ccs:
            msg = "Block Checksum Error! real %02x, calculated %02x" % \
                  (rcs, ccs)
            LOG.error(msg)
            _handshake(radio)
            raise errors.RadioError("Error communicating with radio")

        _handshake(radio, "After checksum in Low Mem")
        return data
    else:
        raise errors.RadioError("Unexpected communication from radio")


def _recvh(radio):
    """Receive high data from the radio, 35 or 4 bytes"""
    rxdata = radio.pipe.read(4)
    # There are two valid options, the first byte is the content
    if (len(rxdata) == 4 and
            rxdata[0] == b"\x5B"[0] and
            rxdata[3] == b"\xFF"[0]):
        # 4 bytes, x5B = empty; payload = xFF (block is all xFF)
        _handshake(radio, "Short block in High Mem")
        return False
    rxdata += radio.pipe.read(RX_BLOCK_SIZE_H - 1)
    if len(rxdata) == RX_BLOCK_SIZE_H + 3 and rxdata[0] == b"\x58"[0]:
        # 35 bytes, x58 + address(2) + data(32), no checksum
        data = rxdata[3:]
        _handshake(radio, "After data in High Mem")
        return data
    else:
        raise errors.RadioError("Unexpected communication from radio")


def _open_radio(radio):
    """Open the radio into program mode and check if it's the correct model"""
    radio.pipe.baudrate = 9600
    radio.pipe.timeout = 1.0

    LOG.debug("Starting program mode.")
    try:
        radio.pipe.write(b"PROGRAM")
    except Exception:
        raise errors.RadioError("Serial Connection Error")

    ack = radio.pipe.read(10)
    if ack == ACK_CMD:
        pass
    elif ack.endswith(b'\xb5\x15\xc5m\xf5\x95\x01') or ack == b'':
        raise errors.RadioError("No response response from radio,"
                                " Is it connected and powered on?")
    else:
        raise errors.RadioError("Radio didn't acknowledge program mode.")

    LOG.debug("Radio entered Program mode.")

    radio.pipe.write(b"\x02\x0F")
    rid = radio.pipe.read(10)

    if not rid.startswith(radio.TYPE):
        LOG.debug("Incorrect model ID:")
        LOG.debug(util.hexprint(rid))
        LOG.debug("expected %s" % radio.TYPE)
        raise errors.RadioError("Radio Model Incorrect")

    LOG.debug("Full radio identity string is:\n%s" % util.hexprint(rid))

    _handshake(radio)


def do_download(radio):
    """ The download function """
    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE
    status.msg = ""
    radio.status_fn(status)

    _open_radio(radio)

    data = b""
    memory_index = 0

    for addr in MEM_LR:
        radio.pipe.write(_make_framel(b"R", addr))
        d = _recvl(radio)
        # if empty block, return false = full of xFF
        if d is False:
            d = EMPTY_L

        data += d

        memory_index += RX_BLOCK_SIZE_L
        status.cur = memory_index
        status.msg = "Cloning from Main MCU (Low mem)..."
        radio.status_fn(status)

    for addr in MEM_MR:
        radio.pipe.write(_make_framem(b"T", addr))
        d = radio.pipe.read(17)

        if len(d) != 17:
            raise errors.RadioError(
                "Problem receiving short block %d on mid mem" % addr)

        data += d[1:]
        _handshake(radio, "Middle mem ack error")

        memory_index += RX_BLOCK_SIZE_M
        status.cur = memory_index
        status.msg = "Cloning from 'unknown' (mid mem)..."
        radio.status_fn(status)

    for addr in MEM_HR:
        radio.pipe.write(_make_frameh(b"S", addr))
        d = _recvh(radio)
        # if empty block, return false = full of xFF
        if d is False:
            d = EMPTY_H

        data += d

        memory_index += RX_BLOCK_SIZE_H
        status.cur = memory_index
        status.msg = "Cloning from Head (High mem)..."
        radio.status_fn(status)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    """ The upload function """
    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE
    status.msg = "Getting the radio into program mode."
    radio.status_fn(status)
    _open_radio(radio)

    memory_index = 0
    img = radio.get_mmap()

    for addr in MEM_LR:
        # this is the data to write
        data = img[memory_index:memory_index + RX_BLOCK_SIZE_L]
        # sdata is the full packet to send

        # flag
        short = False

        # building the data to send
        if data == EMPTY_L:
            # empty block
            sdata = _make_framel(b"Z", addr) + b"\xFF"
            short = True
        else:
            # normal
            cs = _checksum(data)
            sdata = _make_framel(b"W", addr) + data + bytes([cs])

        radio.pipe.write(sdata)

        msg = "Bad ACK on low block %04x" % addr
        _handshake(radio, msg, False)

        memory_index += RX_BLOCK_SIZE_L
        status.cur = memory_index
        status.msg = "Cloning to Main MCU (Low mem)..."
        radio.status_fn(status)

    for addr in MEM_MR:
        data = img[memory_index:memory_index + RX_BLOCK_SIZE_M]
        sdata = _make_framem(b"Y", addr) + b"\x00" + data

        radio.pipe.write(sdata)

        msg = "Bad ACK on mid block %04x" % addr
        _handshake(radio, msg, not short)

        memory_index += RX_BLOCK_SIZE_M
        status.cur = memory_index
        status.msg = "Cloning from middle mem..."
        radio.status_fn(status)

    for addr in MEM_HR:
        data = img[memory_index:memory_index + RX_BLOCK_SIZE_H]
        # sdata is the full packet to send

        # building the data to send
        if data == EMPTY_H:
            # empty block
            sdata = _make_frameh(b"[", addr) + b"\xFF"
        else:
            # normal
            sdata = _make_frameh(b"X", addr) + data

        radio.pipe.write(sdata)

        msg = "Bad ACK on low block %04x" % addr
        _handshake(radio, msg, False)

        memory_index += RX_BLOCK_SIZE_H
        status.cur = memory_index
        status.msg = "Cloning to Head MCU (high mem)..."
        radio.status_fn(status)


def _get_rid(data, index=0x03EE0):
    """Get the radio ID string from a mem string"""
    return data[index:index+6]


def _filter_settings_string(name):
    filtered = ""
    for char in str(name).upper():
        if char in VALID_CHARS:
            filtered += char
        else:
            filtered += " "
    return filtered


class Kenwoodx90BankModel(chirp_common.BankModel):
    """Testing the bank model on kenwood"""
    channelAlwaysHasBank = True

    def get_num_mappings(self):
        return self._radio._num_banks

    def get_mappings(self):
        banks = []
        for i in range(0, self._radio._num_banks):
            bindex = i + 1
            gname = self._radio.get_group_name(i)
            bank = self._radio._bclass(self, bindex, gname)
            bank.index = i
            banks.append(bank)
        return banks

    def add_memory_to_mapping(self, memory, bank):
        self._radio._set_bank(memory.number, bank.index)

    def remove_memory_from_mapping(self, memory, bank):
        if self._radio._get_bank(memory.number) != bank.index:
            raise Exception("Memory %i not in bank %s. Cannot remove." %
                            (memory.number, bank))

        # We can't "Remove" it for good, the kenwood paradigm doesn't allow it
        # instead we move it to bank 0
        self._radio._set_bank(memory.number, 0)

    def get_mapping_memories(self, bank):
        memories = []
        for i in range(0, self._radio._upper):
            if self._radio._get_bank(i) == bank.index:
                memories.append(self._radio.get_memory(i))
        return memories

    def get_memory_mappings(self, memory):
        index = self._radio._get_bank(memory.number)
        if index is None:
            return []
        return [self.get_mappings()[index]]


class MemBank(chirp_common.Bank):
    """A bank model for kenwood"""
    # Integral index of the bank, not to be confused with per-memory
    # bank indexes
    index = 0

    def get_name(self):
        return self._model._radio.get_group_name(self.index)

    def set_name(self, name):
        self._name = self._model._radio.set_group_name(self.index, name)


class Kenwoodx90(chirp_common.CloneModeRadio, chirp_common.ExperimentalRadio):
    """Kenwood TK-790 radio base class"""
    VENDOR = "Kenwood"
    BAUD_RATE = 9600
    VARIANT = ""
    MODEL = ""
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=110),
                    chirp_common.PowerLevel("Low", watts=45)]
    MODES = ["NFM", "FM"]  # 12.5 / 25 Khz
    STEPS = [5.0]
    _name_chars = 8
    _group_name_chars = 8

    _memsize = MEM_SIZE
    _range = [136000000, 162000000]
    _upper = 160
    _banks = None
    _num_banks = 160
    _bclass = MemBank
    _kind = ""
    _head_type = 0
    FORMATS = [directory.register_format('Kenwood KPG-44D', '*.dat')]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            'This driver is experimental: Use at your own risk! '
        rp.pre_download = _(
            "Follow these instructions to download your radio:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio (unlock it if password protected)\n"
            "4 - Click OK to start\n")
        rp.pre_upload = _(
            "Follow these instructions to upload your radio:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio (unlock it if password protected)\n"
            "4 - Click OK to start\n")
        return rp

    def get_features(self):
        """Return information about this radio's features"""
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = True
        rf.has_tuning_step = False
        rf.has_name = True
        rf.has_offset = True
        rf.has_mode = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_modes = self.MODES
        rf.valid_duplexes = ["", "-", "+", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_characters = VALID_CHARS
        rf.valid_skips = SKIP_VALUES
        rf.valid_dtcs_codes = DTCS_CODES
        rf.valid_bands = [self._range]
        rf.valid_name_length = self._name_chars
        rf.valid_tuning_steps = self.STEPS
        rf.memory_bounds = (1, self._upper)
        return rf

    def _fill(self, offset, data):
        """Fill an specified area of the memmap with the passed data"""
        for addr in range(0, len(data)):
            self._mmap[offset + addr] = data[addr]

    def _set_variant(self):
        """Select and set the correct variables for the class according
        to the identified variant of the radio, and other runtime data"""
        rid = _get_rid(self.get_mmap())
        self._banks = dict()
        try:
            self._upper, low, high, self._kind = self.VARIANTS[rid]
            self._range = [low * 1000000, high * 1000000]

        except KeyError:
            LOG.debug("Wrong Kenwood radio, ID or unknown variant")
            LOG.debug(util.hexprint(rid))
            raise errors.RadioError(
                "Wrong Kenwood radio, ID or unknown variant, see LOG output.")

        self._name_chars = int(self._memobj.settings.ch_name_length)
        self._group_name_chars = int(self._memobj.settings.grp_name_length)
        self._head_type = self._memobj.properties.rid.headtype.get_value()

    def sync_in(self):
        """Do a download of the radio eeprom"""
        try:
            self._mmap = do_download(self)
        finally:
            _close_radio(self)
        self.process_mmap()
        self._dat_header_mmap = None

    def sync_out(self):
        """Do an upload to the radio eeprom"""
        try:
            do_upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        finally:
            _close_radio(self)

    def _get_bank_struct(self):
        """Parse the bank data in the mem into the self.bank variable"""
        gl = self._memobj.group_limits
        gb = self._memobj.group_belong
        bank_count = 0

        for bg in gl:
            # check for empty banks
            if bg.index == 255 and bg.length == 255:
                self._banks[bank_count] = list()
                bank_count += 1
                continue

            for i in range(0, bg.length):
                position = bg.index + i
                index = int(gb[position].index)

                try:
                    self._banks[bank_count].append(index)
                except KeyError:
                    self._banks[bank_count] = list()
                    self._banks[bank_count].append(index)

            bank_count += 1

    def process_mmap(self):
        """Process the memory object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

        self._set_variant()

        self._get_bank_struct()

    def load_mmap(self, filename):
        if filename.lower().endswith('.dat'):
            with open(filename, "rb") as f:
                self._dat_header_mmap = memmap.MemoryMapBytes(f.read(0x40))
                self._mmap = memmap.MemoryMapBytes(f.read())
            self.process_mmap()
        else:
            self._dat_header_mmap = None
            chirp_common.CloneModeRadio.load_mmap(self, filename)

    def save_mmap(self, filename):
        if filename.lower().endswith('.dat'):
            with open(filename, 'wb') as f:
                datHeader = self._prep_dat_header()
                f.write(datHeader.get_packed())
                f.write(self._mmap.get_packed())
        else:
            chirp_common.CloneModeRadio.save_mmap(self, filename)

    def _prep_dat_header(self):
        # if we changed head type, the RID changed too
        rid = self._mmap.get(0x3EE0, 10)
        # if dat header imported with file, use it
        if self._dat_header_mmap is not None:
            self._dat_header_mmap.set(0x0F, rid)
            return self._dat_header_mmap
        # otherwise build our own header
        dat_header_map = memmap.MemoryMapBytes(bytes([255]*0x40))
        softwareName = self._mmap.get(0x3EDA, 6)
        softwareVer = self._mmap.get(0x3EFB, 5)
        dat_header_map.set(0x00, softwareName)
        dat_header_map.set(0x0A, softwareVer)
        dat_header_map.set(0x0F, rid)
        return dat_header_map

    def get_raw_memory(self, number):
        """Return a raw representation of the memory object, which
        is very helpful for development"""
        return repr(self._memobj.memory[number])

    def _decode_tone(self, val):
        """Parse the tone data to decode from mem, it returns:
        Mode (''|DTCS|Tone), Value (None|###), Polarity (None,N,R)"""
        val = int(val)
        if val == 65535:
            return '', None, None
        elif val >= 0x2800:
            code = int("%03o" % (val & 0x07FF))
            pol = (val & 0x8000) and "R" or "N"
            return 'DTCS', code, pol
        else:
            a = val / 10.0
            return 'Tone', a, None

    def _encode_tone(self, memval, mode, value, pol):
        """Parse the tone data to encode from UI to mem"""
        if mode == '':
            memval.set_raw(b"\xff\xff")
        elif mode == 'Tone':
            memval.set_value(int(value * 10))
        elif mode == 'DTCS':
            val = int("%i" % value, 8) + 0x2800
            if pol == "R":
                val += 0xA000
            memval.set_value(val)
        else:
            raise Exception("Internal error: invalid mode `%s'" % mode)

    def get_memory(self, number):
        """Get the mem representation from the radio image"""
        _mem = self._memobj.memory[number - 1]

        _chs_names = self._memobj.chs_names[number-1]

        mem = chirp_common.Memory()

        # prevent error when auto-populating new memory (defaults to FM)
        if self.MODEL == "TK-690":
            mem.mode = "NFM"

        mem.number = number

        if _mem.get_raw()[0] == 0xFF:
            mem.empty = True
            return mem

        mem.freq = int(_mem.rxfreq) * 10

        # tx freq can be blank
        if _mem.get_raw()[4] == 0xFF:
            mem.offset = 0
            mem.duplex = "off"
        else:
            offset = (int(_mem.txfreq) * 10) - mem.freq
            if offset < 0:
                mem.offset = abs(offset)
                mem.duplex = "-"
            elif offset > 0:
                mem.offset = offset
                mem.duplex = "+"
            else:
                mem.offset = 0

        mem.name = str(_chs_names.name).rstrip(" ")[:self._name_chars + 1]

        mem.power = self.POWER_LEVELS[int(_mem.power)]

        if self.MODEL == "TK-690" and int(_mem.wide) == 1:
            LOG.debug("Invalid bandwidth mode entry found for TK-690. Fixing")
            _mem.wide = 0
        mem.mode = self.MODES[int(_mem.wide)]

        mem.skip = SKIP_VALUES[int(_mem.add)]

        rxtone = txtone = None
        txtone = self._decode_tone(_mem.txtone)
        rxtone = self._decode_tone(_mem.rxtone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        mem.extra = RadioSettingGroup("extra", "Extra")

        bcl = RadioSetting("bcl", "Busy channel lockout",
                           RadioSettingValueBoolean(bool(_mem.bcl)))
        mem.extra.append(bcl)

        pttid = RadioSetting("pttid", "PTT ID",
                             RadioSettingValueBoolean(bool(_mem.pttid)))
        mem.extra.append(pttid)

        beat = RadioSetting("beatshift", "Beat Shift",
                            RadioSettingValueBoolean(bool(_mem.beatshift)))
        mem.extra.append(beat)

        optsig = RadioSetting("signal", "Optional Signalling",
                              RadioSettingValueList(
                                OPTSIG_LIST, current_index=_mem.signal))
        mem.extra.append(optsig)

        # only TK-790 and TK890 support compander, and only in narrow mode
        if self.MODEL != "TK-690":
            companderDisabled = _mem.wide or _mem.compander
            compander = RadioSetting("compander", "compander",
                                     RadioSettingValueBoolean(
                                         bool(not companderDisabled)))
            mem.extra.append(compander)

        return mem

    def set_memory(self, mem):
        """Set the memory data in the eeprom img from the UI"""
        _mem = self._memobj.memory[mem.number - 1]
        _ch_name = self._memobj.chs_names[mem.number - 1]

        if mem.empty:
            _mem.set_raw(b"\xFF" * 16)

            for byte in _ch_name.name:
                byte.set_raw(b"\xFF")

            self._del_channel_from_bank(mem.number)

            return

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        elif mem.duplex == "off":
            for byte in _mem.txfreq:
                byte.set_raw(b"\xFF")
                _mem.txdisable = 1
        else:
            _mem.txfreq = mem.freq / 10
            _mem.txdisable = 0

        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)
        self._encode_tone(_mem.txtone, txmode, txtone, txpol)
        self._encode_tone(_mem.rxtone, rxmode, rxtone, rxpol)

        _ch_name.name = str(mem.name).ljust(16, " ")

        if mem.power is None:
            _mem.power = 0
        else:
            _mem.power = self.POWER_LEVELS.index(mem.power)

        _mem.wide = self.MODES.index(mem.mode)

        _mem.add = SKIP_VALUES.index(mem.skip)

        # setting required but unknown value
        _mem.valid.set_raw(b'\x00')

        # extra settings
        if len(mem.extra) > 0:
            # there are setting, parse
            for setting in mem.extra:
                if setting.get_name() == "compander":
                    # if width is not Narrow, disable compander
                    if _mem.wide:
                        _mem.compander = True

                    else:
                        _mem.compander = bool(not setting.value)
                else:
                    setattr(_mem, setting.get_name(), setting.value)
        else:
            msg = "Channel #%d has no extra data, loading defaults" % \
                  int(mem.number - 1)
            LOG.info(msg)
            _mem.bcl = 0
            _mem.pttid = 0
            _mem.beatshift = 0
            _mem.compander = 1
            # unknowns
            _mem.signal = 0
            _mem.unknown1 = 0

        # If channel doesn't have any bank assigned, default to bank zero
        b = self._get_bank(mem.number)
        if b is None:
            self._set_bank(mem.number, 0)

        return mem

    @classmethod
    def match_model(cls, filedata, filename):

        # .dat file has constant size and specifies model type near start
        return (filename.lower().endswith('.dat')) and \
           (len(filedata) == DAT_FILE_SIZE) and \
           (_get_rid(filedata, 0x000F) in cls.VARIANTS)

    def get_bank_model(self):
        """Pass the bank model to the UI part"""
        return Kenwoodx90BankModel(self)

    def _get_bank(self, loc):
        """Get the bank data for a specific channel"""
        for k in self._banks:
            if (loc - 1) in self._banks[k]:
                LOG.info("Channel %d is in bank %d" % (loc, k))
                return k
        return None

    def _set_bank(self, loc, bank=0):
        """Set the bank data for a specific channel"""
        # no need to change bank if already set correctly
        b = self._get_bank(loc)
        if b == bank:
            return

        # if another bank already assigned, delete from old location
        if b is not None:
            self._del_channel_from_bank(loc, b)

        self._banks[bank].append(loc - 1)

        self._update_bank_memmap()

    def _del_channel_from_bank(self, loc, bank=None):
        """Remove a channel from a bank, if no bank is specified we find
        where it is"""

        # if we don't know what bank channel is in,
        if bank is None:
            bank = self._get_bank(loc)

        # in case memory entry isn't saved to any bank, no need to delete:
        if bank is not None:
            self._banks[bank].pop(self._banks[bank].index(loc - 1))

        self._update_bank_memmap()

    def _update_bank_memmap(self):
        """This function is called whatever a change is made to a channel
        or a bank, to update the memmap with the changes that was made"""

        bl = b""
        bb = b""

        # group_belong index
        gbi = 0
        # sort the banks for consistient memory entries
        self._banks = dict(sorted(self._banks.items()))
        for bank in self._banks:
            # check for empty banks
            if len(self._banks[bank]) == 0:
                bl += b"\xff\xff"
                continue

            # channel index inside the bank, starting at 1
            # aka channel group index
            cgi = 1
            for channel in range(0, len(self._banks[bank])):
                # update bb
                bb += bytes([cgi, self._banks[bank][channel]])
                # set the group limits for this group
                if cgi == 1:
                    bl += bytes([gbi, len(self._banks[bank])])

                gbi += 1
                cgi += 1

        # fill the gaps before writing
        bb += b"\xff" * 2 * int(self._num_banks - len(bb) / 2)
        bl += b"\xff" * 2 * int(self._num_banks - len(bl) / 2)

        self._fill(0x1480, bl)
        self._fill(0x1600, bb)

    def get_group_name(self, index):
        if self._memobj.grp_names[index].name.get_raw()[0] == (0xFF):
            return ""
        return self._memobj.grp_names[index].name.get_raw()\
            .decode("ascii")[0:self._group_name_chars]

    def set_group_name(self, index, name):
        name = name.upper()
        formatted_name = "".join([x for x in name[:self._group_name_chars]
                                  if x in VALID_CHARS]).ljust(16)[:16]
        self._memobj.grp_names[index].name = formatted_name
        return formatted_name[:self._group_name_chars]

    def get_max_display_length(self):
        max_length = 8
        if self._head_type:  # if full head
            max_length = 14
        return max_length

    def validate_name_lengths(self, new_length):
        max_length = self.get_max_display_length()
        if new_length > max_length:
            return max_length
        return new_length

    def apply_ch_name_length(self, new_length):
        new_length = self.validate_name_lengths(new_length)
        max_length = self.get_max_display_length()
        self._name_chars = new_length
        self._group_name_chars = max_length - new_length
        self._memobj.settings.ch_name_length = new_length
        self._memobj.settings.grp_name_length = max_length - new_length

    def get_settings(self):

        """Translate the MEM_FORMAT structs into the UI"""
        basic_settings = RadioSettingGroup("basic_settings",
                                           "Basic Settings")
        button_assignments = RadioSettingGroup("button_assignments",
                                               "Configurable Button Functions")
        test_frequencies = RadioSettingGroup("test_frequencies",
                                             "Test Frequencies")
        group = RadioSettings(basic_settings, button_assignments,
                              test_frequencies)

        # Basic Settings:
        rs = RadioSetting("head_type",
                          "Type of control-head:",
                          RadioSettingValueList(
                              HEAD_TYPES,
                              current_index=self._memobj.
                              properties.rid.headtype))
        rs.set_volatile(True)
        basic_settings.append(rs)

        ch_len_setting = RadioSettingValueInteger(
            0, 14, self._memobj.settings.ch_name_length)

        rs = RadioSetting("ch_name_length", "Channel Name Length",
                          ch_len_setting)
        rs.set_volatile(True)
        basic_settings.append(rs)

        grp_len_setting = RadioSettingValueInteger(
            0, 14, self._memobj.settings.grp_name_length)
        grp_len_setting.set_mutable(False)

        rs = RadioSetting("grp_name_length", "Group Name Length",
                          grp_len_setting)
        basic_settings.append(rs)

        rs = RadioSetting("poweron_msg", "Power-On Message",
                          RadioSettingValueString(
                            0, 14, _filter_settings_string(
                                self._memobj.poweron_msg),
                            charset=VALID_CHARS_UNCASED))
        basic_settings.append(rs)

        # Button Function Configuration:
        for buttonName in ASSIGNABLE_BUTTONS:
            _fullHeadWarning = ""

            button_func_setting = RadioSettingValueMap(
                    BUTTON_FUNCTION_LIST,
                    self._memobj.button_assignments[buttonName])
            # disable configuring full_head buttons if basic head equipped
            if buttonName in FULL_HEAD_ONLY_BUTTONS and self._head_type == 0:
                _fullHeadWarning = "  (Unavailable on basic control head)"
                button_func_setting.set_mutable(False)

            rs = RadioSetting(
              buttonName,
              "Configured function for "+buttonName+" button"+_fullHeadWarning,
              button_func_setting)
            button_assignments.append(rs)

        for index in range(0, 16):
            rxfreq = int(self._memobj.test_frequencies[index]["rxfreq"])*10
            txfreq = int(self._memobj.test_frequencies[index]["txfreq"])*10
            # don't populate invalid values
            if not ((self._range[0] < rxfreq < self._range[1]) or
                    (self._range[0] < txfreq < self._range[1])):
                rs = RadioSetting("add_testfreq_%i" % index,
                                  "Add test frequency #%i" % index,
                                  RadioSettingValueBoolean(False))
                rs.set_volatile(True)
                test_frequencies.append(rs)
            else:
                for direction in ["rx", "tx"]:
                    key = "set_testfreq_%i_%s" % (index, direction)
                    name = "test frequency %i %s" % (index+1, direction)
                    rs = RadioSetting(key, name,
                                      RadioSettingValueInteger(
                                          self._range[0]/10, self._range[1]/10,
                                          self._memobj.test_frequencies[index]
                                          [direction+"freq"]))
                    test_frequencies.append(rs)
        return group

    def set_settings(self, settings):
        for group in settings:
            for button in group:
                groupKey = group.get_name()
                settingKey = button.get_name()
                if groupKey == "button_assignments":
                    self._memobj[groupKey][settingKey] = \
                        [value for (key, value) in BUTTON_FUNCTION_LIST
                            if key == button.value.get_value()][0]
                elif groupKey == "basic_settings":
                    if settingKey == "head_type":
                        self._memobj.properties.rid.headtype = \
                            int(button.value)
                        self._head_type = int(button.value)
                    elif settingKey == "ch_name_length":
                        self.apply_ch_name_length(int(button.value))
                    elif settingKey == "poweron_msg":
                        self._memobj.poweron_msg = _filter_settings_string(
                            button.value.get_value())
                elif groupKey == "test_frequencies":
                    keyItems = settingKey.split("_")
                    if keyItems[0] == "set":
                        self._memobj.test_frequencies[
                             int(keyItems[2])][
                             keyItems[3]+"freq"] = int(button.value)
                    elif keyItems[0] == "add" and button.value.get_value():
                        centerFreq = (self._range[0]+self._range[1])/2/10
                        self._memobj.test_frequencies[
                             int(keyItems[2])]["rxfreq"] = centerFreq
                        self._memobj.test_frequencies[
                             int(keyItems[2])]["txfreq"] = centerFreq


@directory.register
class TK690Radio(Kenwoodx90):
    """Kenwood TK-690 """
    MODEL = "TK-690"
    TYPE = b"M0690"
    MODES = ["NFM"]
    VARIANTS = {
        b"M0690\x01": (160, 28, 37, "K"),  # see note below
        b"M0690\x02": (160, 35, 43, "K2"),
        b"M0690\x03": (160, 40, 54, "K3")
    }


@directory.register
class TK790Radio(Kenwoodx90):
    """Kenwood TK-790 K/K2"""
    MODEL = "TK-790"
    TYPE = b"M0790"
    VARIANTS = {
        b"M0790\x04": (160, 144, 174, "K"),  # see note below
        b"M0790\x05": (160, 136, 156, "K2")
    }
    STEPS = [5.0, 6.25, 7.5, 12.5]


@directory.register
class TK890Radio(Kenwoodx90):
    """Kenwood TK-890 """
    MODEL = "TK-890"
    TYPE = b"M0890"
    VARIANTS = {
        b"M0890\x06": (160, 450, 490, "K"),
        b"M0890\x07": (160, 480, 512, "K2"),
        b"M0890\x08": (160, 403, 430, "K3"),
        b"M0890\x09": (160, 450, 480, "K(H)")
    }
    STEPS = [5.0, 6.25, 12.5]

    # Note:
    # These radios originally are constrained to certain band segments but the
    # original software doesn't care about it, so in order to match a
    # feature many will miss from the factory software and to help
    # the use of this radios in the ham bands we expanded the range
    # of the "K" version of the TK-790 from 148 to 144, as well as the
    # range of the TK-690 "F1" (from 29.7 to 28) and "F3" (from 50 to 54)
    # versions (note that F3 also needs physical modifications for use with
    # Ham bands)



# Copyright 2012 Dan Smith <dsmith@danplanet.com>
# Copyright 2024 Yuri D'Elia <wavexx@thregr.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, directory, memmap, bitwise, errors
from chirp.settings import (
    RadioSetting, RadioSettings, RadioSettingGroup,
    RadioSettingValueBoolean, RadioSettingValueInteger, RadioSettingValueList
)
import struct

MEM_FORMAT = """
struct {
  u8 _unk1;
  u8 voice: 2,
     beep: 1,
     _unk2: 1,
     vox: 4;
  u8 led_timeout: 4,
     led: 2,
     _unk3: 2;
  u8 _unk4: 4,
     sq: 4;
  u8 _unk5: 2,
     tot: 6;
  u8 _unk6: 4,
     lock_timeout: 4;
  u8 _unk7: 1,
     channel: 7;
  u8 _unk8: 5,
     bat_save: 3;
  u8 _unk9[2];
  u8 pass[6];
} settings;

struct {
  u8 freq[5]; // 20 bit rx + 20 bit tx
  u16 low_pwr: 1,
      no_tx: 1,
      rx_tone: 2,
      rx_code: 12;
  u16 nfm: 1,
      skip: 1,
      tx_tone: 2,
      tx_code: 12;
  u8 compander: 2, // only 0b00/0b01 allowed
     hopping: 2, // only 0b00/0b01 allowed
     scrambler: 4;
} memory[80];
"""

VOICE_LIST = ["off", "Chinese", "English"]
SCRAMBLER_LIST = ["off", "1", "2", "3", "4", "5", "6", "7", "8"]
LED_LIST = ["Low", "Medium", "High"]
BAT_SAVE_LIST = ["off", "1:1", "1:2", "1:3", "1:4"]
TONE_LIST = ["Tone", "DTCS_N", "DTCS_I", ""]
LED_TIMEOUT_LIST = ["Continuous", "5", "10", "15", "20", "25", "30",
                    "35", "40", "45", "50", "55", "60"]
LOCK_TIMEOUT_LIST = ["off", "5", "10", "15", "20", "25", "30",
                     "35", "40", "45", "50", "55", "60"]
TOT_LIST = ["off", "15", "30", "45", "60", "75", "90",
            "105", "120", "135", "150", "165", "180", "195",
            "210", "225", "240", "255", "270", "285",
            "300", "315", "330", "345", "360", "375", "390",
            "405", "420", "435", "450", "465", "480", "495",
            "510", "525", "540", "555", "570", "585", "600"]
POWER_LIST = [chirp_common.PowerLevel("High", watts=2.00),
              chirp_common.PowerLevel("Low",  watts=0.50)]


def _checksum(data):
    cs = 2
    for byte in data:
        cs += byte
    return cs % 256


def enter_programming_mode(radio):
    serial = radio.pipe

    cmd = b"\x32\x31\x05\x10"
    req = cmd + bytes([_checksum(cmd)])

    try:
        serial.write(req)
        res = serial.read(1)
        if res != b"\x06":
            raise Exception("invalid response")
    except Exception as e:
        msg = "Radio refused to enter programming mode: %s" % str(e)
        raise errors.RadioError(msg)


def exit_programming_mode(radio):
    serial = radio.pipe

    cmd = b"\x32\x31\x05\xee"
    req = cmd + bytes([_checksum(cmd)])

    try:
        # there is no response from this command as the radio resets
        serial.write(req)
    except Exception:
        raise errors.RadioError("Radio refused to exit programming mode")


def _read_block(radio, block_addr):
    serial = radio.pipe

    cmd = struct.pack(">cH", b"R", block_addr)
    req = cmd + bytes([_checksum(cmd)])

    try:
        serial.write(req)
        res_len = len(cmd) + radio.BLOCK_SIZE + 1
        res = serial.read(res_len)

        if len(res) != res_len or res[:len(cmd)] != cmd:
            raise Exception("unexpected reply!")
        if res[-1] != _checksum(res[:-1]):
            raise Exception("block failed checksum!")

        block_data = res[len(cmd):-1]
    except Exception as e:
        msg = "Failed to read block at %04x: %s" % (block_addr, str(e))
        raise errors.RadioError(msg)

    return block_data


def _write_block(radio, block_addr, block_data):
    serial = radio.pipe

    cmd = struct.pack(">cH", b"W", block_addr) + block_data
    req = cmd + bytes([_checksum(cmd)])

    try:
        serial.write(req)
        res = serial.read(1)
        if res != b"\x06":
            raise Exception("unexpected reply!")
    except Exception as e:
        msg = "Failed to write block at %04x: %s" % (block_addr, str(e))
        raise errors.RadioError(msg)


def verify_model(radio):
    # Simply rely on the protocol/checksum to validate the radio model
    # for now: attempt at least twice, so that garbage in the line is
    # ignored on the first tries
    for _ in range(3):
        try:
            _read_block(radio, radio.START_ADDR)
            return
        except Exception:
            pass

    raise errors.RadioError("Could not communicate with the radio")


def do_download(radio):
    status = chirp_common.Status()
    status.msg = "Cloning from radio"
    status.max = radio._memsize

    verify_model(radio)

    data = b""
    for addr in range(radio.START_ADDR,
                      radio.START_ADDR + radio._memsize,
                      radio.BLOCK_SIZE):
        status.cur = addr
        radio.status_fn(status)

        block = _read_block(radio, addr)
        data += block

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    verify_model(radio)
    enter_programming_mode(radio)

    status = chirp_common.Status()
    status.msg = "Uploading to radio"
    status.max = radio._memsize
    mmap = radio.get_mmap()

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr
        radio.status_fn(status)

        block = mmap[addr:addr + radio.BLOCK_SIZE]
        _write_block(radio, radio.START_ADDR + addr, block)

    exit_programming_mode(radio)


def mem_to_triplet(mem_tone, mem_code):
    mem_tone = TONE_LIST[mem_tone]
    if mem_tone == "Tone":
        mode = "Tone"
        code = mem_code / 10
        polarity = None
    elif mem_tone in ["DTCS_N", "DTCS_I"]:
        mode = "DTCS"
        code = int("%o" % mem_code)
        polarity = "N" if mem_tone == "DTCS_N" else "R"
    else:
        mode = None
        code = None
        polarity = None
    return (mode, code, polarity)


def triplet_to_mem(tone):
    mode, code, polarity = tone
    if mode == "Tone":
        mem_tone = "Tone"
        mem_code = int(code * 10)
    elif mode == "DTCS":
        mem_tone = "DTCS_N" if polarity == "N" else "DTCS_I"
        mem_code = int('%i' % code, 8)
    else:
        mem_tone = ""
        mem_code = 0
    mem_tone = TONE_LIST.index(mem_tone)
    return (mem_tone, mem_code)


@directory.register
class KSunM6Radio(chirp_common.CloneModeRadio):
    VENDOR = "KSUN"
    MODEL = "M6"
    BAUD_RATE = 4800
    BLOCK_SIZE = 0x10
    START_ADDR = 0x0050
    CHANNELS = 80
    _memsize = BLOCK_SIZE + 10 * CHANNELS

    # Return information about this radio's features, including
    # how many memories it has, what bands it supports, etc
    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_bank = False
        rf.has_name = False
        rf.has_settings = True
        rf.memory_bounds = (1, self.CHANNELS)

        rf.can_odd_split = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.valid_duplexes = ["", "split", "off"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "DTCS->", "->DTCS", "Tone->DTCS",
                                "DTCS->Tone", "->Tone", "DTCS->DTCS"]

        rf.has_tuning_step = False
        rf.has_nostep_tuning = True
        rf.valid_bands = [(400000000, 480000000)]

        rf.valid_modes = ["FM", "NFM"]
        rf.valid_power_levels = POWER_LIST

        return rf

    def get_settings(self):
        settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        top = RadioSettings(basic)

        voice = settings.voice
        rsv = RadioSettingValueList(VOICE_LIST, current_index=voice)
        rs = RadioSetting("voice", "Voice language", rsv)
        basic.append(rs)

        led = settings.led
        rsv = RadioSettingValueList(LED_LIST, current_index=led)
        rs = RadioSetting("led", "LED brightness", rsv)
        basic.append(rs)

        led_timeout = settings.led_timeout
        rsv = RadioSettingValueList(LED_TIMEOUT_LIST,
                                    current_index=led_timeout)
        rs = RadioSetting("led_timeout", "LED timeout", rsv)
        basic.append(rs)

        lock_timeout = settings.lock_timeout
        rsv = RadioSettingValueList(LOCK_TIMEOUT_LIST,
                                    current_index=lock_timeout)
        rs = RadioSetting("lock_timeout", "Key Lock timeout", rsv)
        basic.append(rs)

        tot = settings.tot
        rsv = RadioSettingValueList(TOT_LIST, current_index=tot)
        rs = RadioSetting("tot", "Time-Out Timer", rsv)
        basic.append(rs)

        bat_save = settings.bat_save
        rsv = RadioSettingValueList(BAT_SAVE_LIST, current_index=bat_save)
        rs = RadioSetting("bat_save", "Battery Save", rsv)
        basic.append(rs)

        rsv = RadioSettingValueInteger(0, 9, settings.sq)
        rs = RadioSetting("sq", "Squelch Level", rsv)
        basic.append(rs)

        rsv = RadioSettingValueInteger(0, 9, settings.vox)
        rs = RadioSetting("vox", "VOX Level", rsv)
        basic.append(rs)

        rsv = RadioSettingValueBoolean(settings.beep)
        rs = RadioSetting("beep", "Beep", rsv)
        basic.append(rs)

        channel = settings.channel + 1
        rsv = RadioSettingValueInteger(1, self.CHANNELS, channel)
        rs = RadioSetting("channel", "Current Channel", rsv)
        basic.append(rs)

        return top

    def set_settings(self, settings):
        settings = settings[0]
        _settings = self._memobj.settings
        _settings.voice = VOICE_LIST.index(settings["voice"].value.get_value())
        _settings.led = LED_LIST.index(settings["led"].value.get_value())
        _settings.led_timeout = LED_TIMEOUT_LIST.index(
            settings["led_timeout"].value.get_value())
        _settings.lock_timeout = LOCK_TIMEOUT_LIST.index(
            settings["lock_timeout"].value.get_value())
        _settings.tot = TOT_LIST.index(settings["tot"].value.get_value())
        _settings.bat_save = BAT_SAVE_LIST.index(
            settings["bat_save"].value.get_value())
        _settings.sq = settings["sq"].value.get_value()
        _settings.vox = settings["vox"].value.get_value()
        _settings.beep = settings["beep"].value.get_value()
        _settings.channel = settings["channel"].value.get_value() - 1

    # Do a download of the radio from the serial port
    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    # Do an upload of the radio to the serial port
    def sync_out(self):
        do_upload(self)

    # Convert the raw byte array into a memory object structure
    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    # Return a raw representation of the memory object, which
    # is very helpful for development
    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number-1])

    # Extract a high-level memory object from the low-level memory map
    # This is called to populate a memory in the UI
    def get_memory(self, number):
        _mem = self._memobj.memory[number-1]

        mem = chirp_common.Memory()
        mem.number = number

        if _mem.freq.get_raw() == bytes([255] * 5):
            mem.empty = True
        else:
            rx_freq = ((_mem.freq[0] << 12)
                       + (_mem.freq[1] << 4) + (_mem.freq[2] >> 4))
            tx_freq = (((_mem.freq[2] & 0xF) << 16)
                       + (_mem.freq[3] << 8) + _mem.freq[4])

            mem.freq = rx_freq * 1000000 // 2000
            mem.offset = tx_freq * 1000000 // 2000

            if _mem.no_tx:
                mem.duplex = "off"
            elif rx_freq != tx_freq:
                mem.duplex = "split"

            chirp_common.split_tone_decode(
                mem,
                mem_to_triplet(_mem.tx_tone, _mem.tx_code),
                mem_to_triplet(_mem.rx_tone, _mem.rx_code))

        mem.mode = "NFM" if not mem.empty and _mem.nfm else "FM"
        mem.power = POWER_LIST[int(not mem.empty and _mem.low_pwr)]
        mem.skip = "S" if not mem.empty and _mem.skip else ""

        mem.extra = RadioSettingGroup("Extra", "extra")

        hopping = False if mem.empty else _mem.hopping
        rsv = RadioSettingValueBoolean(hopping)
        rs = RadioSetting("hopping", "Hopping", rsv)
        mem.extra.append(rs)

        compander = False if mem.empty else _mem.compander
        rsv = RadioSettingValueBoolean(compander)
        rs = RadioSetting("compander", "Compander", rsv)
        mem.extra.append(rs)

        scrambler = False if mem.empty else _mem.scrambler
        rsv = RadioSettingValueList(SCRAMBLER_LIST, current_index=scrambler)
        rs = RadioSetting("scrambler", _("Scrambler"), rsv)
        mem.extra.append(rs)

        return mem

    # Store details about a high-level memory to the memory map
    # This is called when a user edits a memory in the UI
    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number-1]

        if mem.empty:
            _mem.fill_raw(b"\xff")
        else:
            rx_freq = mem.freq
            if mem.duplex == "split" and mem.offset:
                tx_freq = mem.offset
            else:
                tx_freq = rx_freq

            rx_freq = round(rx_freq / 1000000 * 2000)
            tx_freq = round(tx_freq / 1000000 * 2000)

            _mem.freq[0] = (rx_freq >> 12) & 0xff
            _mem.freq[1] = (rx_freq >> 4) & 0xff
            _mem.freq[2] = ((rx_freq << 4) & 0xf0) | ((tx_freq >> 16) & 0x0f)
            _mem.freq[3] = (tx_freq >> 8) & 0xff
            _mem.freq[4] = (tx_freq) & 0xff

            tx_tone, rx_tone = chirp_common.split_tone_encode(mem)
            _mem.tx_tone, _mem.tx_code = triplet_to_mem(tx_tone)
            _mem.rx_tone, _mem.rx_code = triplet_to_mem(rx_tone)

            _mem.no_tx = (mem.duplex == "off")
            _mem.nfm = (mem.mode == "NFM")
            _mem.skip = (mem.skip == "S")

            if mem.power in POWER_LIST:
                _mem.low_pwr = POWER_LIST.index(mem.power)
            else:
                _mem.low_pwr = False

            if "hopping" in mem.extra:
                _mem.hopping = mem.extra["hopping"].value.get_value()
            else:
                _mem.hopping = False

            if "compander" in mem.extra:
                _mem.compander = mem.extra["compander"].value.get_value()
            else:
                _mem.compander = False

            if "scrambler" in mem.extra:
                _mem.scrambler = SCRAMBLER_LIST.index(
                    mem.extra["scrambler"].value.get_value())
            else:
                _mem.scrambler = 0



import os
import struct
import unittest

from chirp.drivers import kguv8e
from chirp.drivers import kguv8d
from chirp.drivers import kguv8dplus
from chirp.drivers import kg935g
from chirp import directory


class FakeUV8D:
    def __init__(self, rclass):
        self.writebuf = b''
        self.readbuf = b''
        self.rclass = rclass
        self._model = rclass._model

    def write(self, data):
        assert isinstance(data, bytes)
        cmd = data[1]
        response = b''
        if cmd == kguv8d.CMD_ID:
            response = self._model + b'\x00' * 10
        elif cmd == kguv8d.CMD_RD:
            response = b'\x00' * 64
        elif cmd == kguv8d.CMD_WR:
            response = data[4:6]
            if hasattr(self.rclass, 'decrypt'):
                response = self.rclass(None).decrypt(response)
        elif cmd == kguv8d.CMD_END:
            return
        else:
            raise Exception('Unhandled command')
        packet = struct.pack('xxxB', len(response)) + response
        cs = sum(x for x in packet) % 256
        packet += struct.pack('B', cs)
        if hasattr(self.rclass, 'encrypt'):
            payload = self.rclass(None).encrypt(packet[4:])
            packet = packet[:4] + payload

        self.readbuf += packet

    def read(self, n):
        buf = self.readbuf[:n]
        self.readbuf = self.readbuf[n:]
        return buf


class TestKGUV8D(unittest.TestCase):
    RCLASS = kguv8d.KGUV8DRadio

    def test_identify(self):
        f = FakeUV8D(self.RCLASS)
        r = self.RCLASS(f)
        r._identify()

    def test_download(self):
        f = FakeUV8D(self.RCLASS)
        r = self.RCLASS(f)
        r.sync_in()

    def test_upload(self):
        f = FakeUV8D(self.RCLASS)
        img = os.path.join(os.path.dirname(__file__),
                           '..', 'images',
                           '%s.img' % directory.radio_class_id(self.RCLASS))
        r = self.RCLASS(img)
        r.set_pipe(f)
        r.sync_out()


class TestKGUV8DPlus(TestKGUV8D):
    RCLASS = kguv8dplus.KGUV8DPlusRadio


class TestKGUV8ER(TestKGUV8D):
    RCLASS = kguv8e.KGUV8ERadio


class TestKG935(TestKGUV8D):
    RCLASS = kg935g.KG935GRadio



from chirp import chirp_common
from chirp import settings
from tests import base


class TestCaseSettings(base.DriverTest):
    def test_has_settings(self):
        return
        settings = self.radio.get_settings()
        if settings:
            self.assertFalse(self.rf.has_settings,
                             'Radio returned settings but has_settings=False')
        else:
            self.assertTrue(self.rf.has_settings,
                            'Radio returned no settings but has_settings=True')

    @base.requires_feature('has_settings')
    def test_get_settings(self):
        lst = self.radio.get_settings()
        self.assertIsInstance(lst, list)

    @base.requires_feature('has_settings')
    def test_same_settings(self):
        o = self.radio.get_settings()
        self.radio.set_settings(o)
        n = self.radio.get_settings()
        list(map(self.compare_settings, o, n))

    def compare_settings(self, a, b):
        try:
            if isinstance(a, settings.RadioSettingValue):
                raise StopIteration
            list(map(self.compare_settings, a, b))
        except StopIteration:
            self.assertEqual(a.get_value(), b.get_value(),
                             'Setting value changed from %r to %r' % (
                                 a.get_value(), b.get_value()))
        except Exception as e:
            print('Failed %s on %s' % (e, a.get_name()))
            raise

    def test_memory_extra_frozen(self):
        # Find the first non-empty memory and try to set it back as a
        # FrozenMemory to make sure the driver does not try to modify
        # any of the settings.
        for i in range(*self.rf.memory_bounds):
            m = self.radio.get_memory(i)
            if not m.empty:
                self.radio.set_memory(chirp_common.FrozenMemory(m))
                break

    def test_memory_extra_flat(self):
        for i in range(*self.rf.memory_bounds):
            m = self.radio.get_memory(i)
            if not m.empty:
                self.assertIsInstance(
                    m.extra,
                    (list, settings.RadioSettingGroup),
                    'mem.extra must be a list or RadioSettingGroup')
                for e in m.extra:
                    self.assertIsInstance(
                        e, settings.RadioSetting,
                        'mem.extra items must be RadioSetting objects')



# Copyright 2013 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tests.unit import base
from chirp import settings


class TestSettingValues(base.BaseTest):
    def _set_and_test(self, rsv, *values):
        for value in values:
            rsv.set_value(value)
            self.assertEqual(rsv.get_value(), value)

    def _set_and_catch(self, rsv, *values):
        for value in values:
            self.assertRaises(settings.InvalidValueError,
                              rsv.set_value, value)

    def test_radio_setting_value_integer(self):
        value = settings.RadioSettingValueInteger(0, 10, 5)
        value.initialize()
        self.assertEqual(value.get_value(), 5)
        self._set_and_test(value, 1, 0, 10)
        self._set_and_catch(value, -1, 11)

    def test_radio_setting_value_float(self):
        value = settings.RadioSettingValueFloat(1.0, 10.5, 5.0)
        value.initialize()
        self.assertEqual(value.get_value(), 5.0)
        self._set_and_test(value, 2.5, 1.0, 10.5)
        self._set_and_catch(value, 0.9, 10.6, -1.5)

    def test_radio_setting_value_boolean(self):
        value = settings.RadioSettingValueBoolean(True)
        value.initialize()
        self.assertTrue(value.get_value())
        self._set_and_test(value, True, False)

    def test_radio_setting_value_list(self):
        opts = ["Abc", "Def", "Ghi"]
        value = settings.RadioSettingValueList(opts, "Abc")
        value.initialize()
        self.assertEqual(value.get_value(), "Abc")
        self.assertEqual(int(value), 0)
        self._set_and_test(value, "Def", "Ghi", "Abc")
        self._set_and_catch(value, "Jkl", "Xyz")
        self.assertEqual(value.get_options(), opts)

    def test_radio_setting_value_list_by_index(self):
        opts = ["Abc", "Def", "Ghi"]
        value = settings.RadioSettingValueList(opts, current_index=1)
        value.initialize()
        self.assertEqual(value.get_value(), "Def")
        self.assertEqual(int(value), 1)
        self._set_and_test(value, "Def", "Ghi", "Abc")
        self._set_and_catch(value, "Jkl", "Xyz")
        self.assertEqual(value.get_options(), opts)

        # Make sure we int() the index, as we may pass a bitwise object
        value = settings.RadioSettingValueList(opts, current_index='1')
        value.initialize()
        self.assertEqual(value.get_value(), "Def")

        value.set_index(0)
        self.assertEqual(value.get_value(), 'Abc')

        # Same here, int() the index
        value.set_index('0')
        self.assertEqual(value.get_value(), 'Abc')

        self.assertRaises(IndexError, value.set_index, 7)

    def test_radio_setting_value_string(self):
        value = settings.RadioSettingValueString(1, 5, "foo", autopad=False)
        value.initialize()
        self.assertEqual(value.get_value(), "foo")
        self.assertEqual(str(value), "foo")
        self._set_and_test(value, "a", "abc", "abdef")
        self._set_and_catch(value, "", "abcdefg")

    def test_validate_callback(self):
        class TestException(Exception):
            pass

        value = settings.RadioSettingValueString(0, 5, "foo", autopad=False)
        value.initialize()

        def test_validate(val):
            if val == "bar":
                raise TestException()
        value.set_validate_callback(test_validate)
        value.set_value("baz")
        self.assertRaises(TestException, value.set_value, "bar")

    def test_changed(self):
        value = settings.RadioSettingValueBoolean(False)
        value.initialize()
        self.assertFalse(value.changed())
        value.set_value(False)
        self.assertFalse(value.changed())
        value.set_value(True)
        self.assertTrue(value.changed())


class TestSettingContainers(base.BaseTest):
    def test_radio_setting_group(self):
        s1 = settings.RadioSetting("s1", "Setting 1")
        s2 = settings.RadioSetting("s2", "Setting 2")
        s3 = settings.RadioSetting("s3", "Setting 3")
        group = settings.RadioSettingGroup("foo", "Foo Group", s1)
        self.assertEqual(group.get_name(), "foo")
        self.assertEqual(group.get_shortname(), "Foo Group")
        self.assertEqual(group.values(), [s1])
        self.assertEqual(group.keys(), ["s1"])
        group.append(s2)
        self.assertEqual(group.items(), [("s1", s1), ("s2", s2)])
        self.assertEqual(group["s1"], s1)
        group["s3"] = s3
        self.assertEqual(group.values(), [s1, s2, s3])
        self.assertEqual(group.keys(), ["s1", "s2", "s3"])
        self.assertEqual([x for x in group], [s1, s2, s3])

        def set_dupe():
            group["s3"] = s3
        self.assertRaises(KeyError, set_dupe)

    def test_radio_setting(self):
        val = settings.RadioSettingValueBoolean(True)
        rs = settings.RadioSetting("foo", "Foo", val)
        self.assertEqual(rs.value, val)
        rs.value = False
        self.assertEqual(val.get_value(), False)
        self.assertEqual('foo:False', str(rs))
        self.assertEqual('[RadioSetting foo:False]', repr(rs))

    def test_radio_setting_multi(self):
        val1 = settings.RadioSettingValueBoolean(True)
        val2 = settings.RadioSettingValueBoolean(False)
        rs = settings.RadioSetting("foo", "Foo", val1, val2)
        self.assertEqual(rs[0], val1)
        self.assertEqual(rs[1], val2)
        rs[0] = False
        rs[1] = True
        self.assertEqual(val1.get_value(), False)
        self.assertEqual(val2.get_value(), True)

    def test_apply_callback(self):
        class TestException(Exception):
            pass

        rs = settings.RadioSetting("foo", "Foo")
        self.assertFalse(rs.has_apply_callback())

        def test_cb(setting, data1, data2):
            self.assertEqual(setting, rs)
            self.assertEqual(data1, "foo")
            self.assertEqual(data2, "bar")
            raise TestException()
        rs.set_apply_callback(test_cb, "foo", "bar")
        self.assertTrue(rs.has_apply_callback())
        self.assertRaises(TestException, rs.run_apply_callback)

    def test_setting_banned_name(self):
        self.assertRaises(settings.InvalidNameError,
                          settings.RadioSetting, "foo%bar", "Foo")

    def test_setting_banned_name_characters(self):
        for c in settings.BANNED_NAME_CHARACTERS:
            self.assertRaises(settings.InvalidNameError,
                              settings.RadioSetting, "foo%sbar" % c, "Foo")

    def test_setting_deferred(self):
        val = settings.RadioSettingValueInteger(0, 10, 12)
        rs = settings.RadioSetting('test', 'Test', val)
        self.assertFalse(rs.value.initialized)
        rs.value = 1
        self.assertTrue(rs.value.initialized)
